"""Upgrade a hardware server."""
# :license: MIT, see LICENSE for more details.

import click
from SoftLayer import utils

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--memory', type=click.INT, help="Memory Size in GB")
@click.option('--network', help="Network port speed in Mbps",
              default=None,
              type=click.Choice(['100', '100 Redundant', '100 Dual',
                                 '1000', '1000 Redundant', '1000 Dual',
                                 '10000', '10000 Redundant', '10000 Dual'])
              )
@click.option('--drive-controller',
              help="Drive Controller",
              default=None,
              type=click.Choice(['Non-RAID', 'RAID']))
@click.option('--public-bandwidth', type=click.INT, help="Public Bandwidth in GB")
@click.option('--add-disk', nargs=2, multiple=True, type=(int, int),
              help="Add a Hard disk in GB to a specific channel, e.g 1000 GB in disk2, it will be "
                   "--add-disk 1000 2")
@click.option('--resize-disk', nargs=2, multiple=True, type=(int, int),
              help="Upgrade a specific disk size in GB, e.g --resize-disk 2000 2")
@click.option('--test', is_flag=True, default=False, help="Do not actually upgrade the hardware server")
@environment.pass_env
def cli(env, identifier, memory, network, drive_controller, public_bandwidth, add_disk, resize_disk, test):
    """Upgrade a Hardware Server."""

    mgr = SoftLayer.HardwareManager(env.client)
    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    if not any([memory, network, drive_controller, public_bandwidth, add_disk, resize_disk]):
        raise exceptions.ArgumentError("Must provide "
                                       " [--memory], [--network], [--drive-controller], [--public-bandwidth],"
                                       "[--add-disk] or [--resize-disk]")

    hw_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'Hardware')
    if not test:
        if not (env.skip_confirmations or formatting.confirm(
                "This action will incur charges on your account. Continue?")):
            raise exceptions.CLIAbort('Aborted')

    disk_list = []
    if add_disk:
        for guest_disk in add_disk:
            disks = {'description': 'add_disk', 'capacity': guest_disk[0], 'number': guest_disk[1]}
            disk_list.append(disks)
    if resize_disk:
        for guest_disk in resize_disk:
            disks = {'description': 'resize_disk', 'capacity': guest_disk[0], 'number': guest_disk[1]}
            disk_list.append(disks)

    response = mgr.upgrade(hw_id, memory=memory, nic_speed=network, drive_controller=drive_controller,
                           public_bandwidth=public_bandwidth, disk=disk_list, test=test)

    if response:
        if test:
            add_data_to_table(response, table)
        else:
            table.add_row(['Order Date', response.get('orderDate')])
            table.add_row(['Order Id', response.get('orderId')])
            add_data_to_table(response['orderDetails'], table)
            place_order_table = get_place_order_information(response)
            table.add_row(['Place Order Information', place_order_table])
            order_detail_table = get_order_detail(response)
            table.add_row(['Order Detail (Billing Information)', order_detail_table])

    env.fout(table)


def add_data_to_table(response, table):
    """Add the hardware server upgrade result to the table"""
    table.add_row(['location', utils.lookup(response, 'locationObject', 'longName')])
    table.add_row(['quantity', response.get('quantity')])
    table.add_row(['Package Id', response.get('packageId')])
    table.add_row(['Currency Short Name', response.get('currencyShortName')])
    table.add_row(['Prorated Initial Charge', response.get('proratedInitialCharge')])
    table.add_row(['Prorated Order Total', response.get('proratedOrderTotal')])
    table.add_row(['Hourly Pricing', response.get('useHourlyPricing')])
    table_hardware = get_hardware_detail(response)
    table.add_row(['Hardware', table_hardware])
    table_prices = get_hardware_prices(response)
    table.add_row(['Prices', table_prices])


def get_place_order_information(response):
    """Get the hardware server place order information."""
    table_place_order = formatting.Table(['Id', 'Account Id', 'Status', 'Account CompanyName',
                                          'UserRecord FirstName', 'UserRecord LastName', 'UserRecord Username'])
    table_place_order.add_row([response.get('id'),
                               response.get('accountId'),
                               response.get('status'),
                               utils.lookup(response, 'account', 'companyName'),
                               utils.lookup(response, 'userRecord', 'firstName'),
                               utils.lookup(response, 'account', 'lastName'),
                               utils.lookup(response, 'account', 'username')])

    return table_place_order


def get_hardware_detail(response):
    """Get the hardware server detail."""
    table_hardware = formatting.Table(['Account Id', 'Hostname', 'Domain'])
    for hardware in response['hardware']:
        table_hardware.add_row([hardware.get('accountId'),
                                hardware.get('hostname'),
                                hardware.get('domain')])

    return table_hardware


def get_hardware_prices(response):
    """Get the hardware server prices."""
    table_prices = formatting.Table(['Id', 'HourlyRecurringFee', 'RecurringFee', 'Categories', 'Item Description',
                                     'Item Units'])
    for price in response['prices']:
        categories = price.get('categories')[0]
        table_prices.add_row([price.get('id'),
                              price.get('hourlyRecurringFee'),
                              price.get('recurringFee'),
                              categories.get('name'),
                              utils.lookup(price, 'item', 'description'),
                              utils.lookup(price, 'item', 'units')])

    return table_prices


def get_order_detail(response):
    """Get the hardware server order detail."""
    table_order_detail = formatting.Table(['Billing City', 'Billing Country Code', 'Billing Email',
                                           'Billing Name First', 'Billing Name Last', 'Billing Postal Code',
                                           'Billing State'])
    table_order_detail.add_row([utils.lookup(response, 'orderDetails', 'billingInformation', 'billingCity'),
                                utils.lookup(response, 'orderDetails', 'billingInformation', 'billingCountryCode'),
                                utils.lookup(response, 'orderDetails', 'billingInformation', 'billingEmail'),
                                utils.lookup(response, 'orderDetails', 'billingInformation', 'billingNameFirst'),
                                utils.lookup(response, 'orderDetails', 'billingInformation', 'billingNameLast'),
                                utils.lookup(response, 'orderDetails', 'billingInformation', 'billingPostalCode'),
                                utils.lookup(response, 'orderDetails', 'billingInformation', 'billingState')])

    return table_order_detail



"""Share an image template with another account."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--account-id', help='Account Id for another account to share image template', required=True)
@environment.pass_env
def cli(env, identifier, account_id):
    """Share an image template with another account."""

    image_mgr = SoftLayer.ImageManager(env.client)
    image_id = helpers.resolve_id(image_mgr.resolve_ids, identifier, 'image')
    shared_image = image_mgr.share_image(image_id, account_id)

    if shared_image:
        env.fout(f"Image template {identifier} was shared to account {account_id}.")



"""Toggle the IPMI interface on and off."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--enable/--disable', default=True,
              help="Whether enable (DEFAULT) or disable the interface.")
@environment.pass_env
def cli(env, identifier, enable):
    """Toggle the IPMI interface on and off"""

    mgr = SoftLayer.HardwareManager(env.client)
    hw_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'hardware')
    result = env.client['Hardware_Server'].toggleManagementInterface(enable, id=hw_id)
    env.fout(result)



"""Verify or place an order."""
# :license: MIT, see LICENSE for more details.

import json

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.managers import NetworkManager
from SoftLayer.managers import ordering


COLUMNS = ['keyName', 'description', 'cost']


@click.command(cls=SLCommand)
@click.argument('package_keyname')
@click.argument('location')
@click.option('--preset',
              help="The order preset (if required by the package)")
@click.option('--verify', is_flag=True,
              help="Flag denoting whether or not to only verify the order, not place it")
@click.option('--quantity', type=int, default=1,
              help="The quantity of the item being ordered")
@click.option('--billing', type=click.Choice(['hourly', 'monthly']), default='hourly', show_default=True,
              help="Billing rate")
@click.option('--complex-type',
              help="The complex type of the order. Starts with 'SoftLayer_Container_Product_Order'.")
@click.option('--extras',
              help="JSON string denoting extra data that needs to be sent with the order")
@click.argument('order_items', nargs=-1)
@environment.pass_env
def cli(env, package_keyname, location, preset, verify, billing, complex_type,
        quantity, extras, order_items):
    """Place or verify an order.

\b
1. Find the package keyName from `slcli order package-list`
2. Find the location from `slcli order package-locations PUBLIC_CLOUD_SERVER`
  If the package does not require a location, use 'NONE' instead.
3. Find the needed items `slcli order item-list PUBLIC_CLOUD_SERVER`
  Some packages, like PUBLIC_CLOUD_SERVER need presets, `slcli order preset-list PUBLIC_CLOUD_SERVER`
4. Find the complex type from https://sldn.softlayer.com/reference
5. Use that complex type to fill out any --extras

    Example::

        slcli order place --verify --preset B1_2X8X100 --billing hourly
        --complex-type SoftLayer_Container_Product_Order_Virtual_Guest
        --extras '{"virtualGuests": [{"hostname": "test", "domain": "ibm.com"}]}'
        PUBLIC_CLOUD_SERVER DALLAS13
        BANDWIDTH_0_GB_2 MONITORING_HOST_PING NOTIFICATION_EMAIL_AND_TICKET
        OS_DEBIAN_9_X_STRETCH_LAMP_64_BIT 1_IP_ADDRESS 1_IPV6_ADDRESS
        1_GBPS_PUBLIC_PRIVATE_NETWORK_UPLINKS REBOOT_REMOTE_CONSOLE
        AUTOMATED_NOTIFICATION UNLIMITED_SSL_VPN_USERS_1_PPTP_VPN_USER_PER_ACCOUNT

    """
    manager = ordering.OrderingManager(env.client)
    network = NetworkManager(env.client)

    # Check if this location is going to be shutdown soon.
    pods = network.get_closed_pods()
    location_dc = network.get_datacenter_by_keyname(location)
    for pod in pods:
        if location_dc and location_dc.get('name') in pod.get('name'):
            click.secho(f"Warning: Closed soon: {pod.get('name')}", fg='yellow')

    if extras:
        try:
            extras = json.loads(extras)
        except ValueError as err:
            raise exceptions.CLIAbort(f"There was an error when parsing the --extras value: {err}")

    args = (package_keyname, location, order_items)
    kwargs = {'preset_keyname': preset,
              'extras': extras,
              'quantity': quantity,
              'complex_type': complex_type,
              'hourly': bool(billing == 'hourly')}

    if verify:
        table = formatting.Table(COLUMNS)
        order_to_place = manager.verify_order(*args, **kwargs)
        for price in order_to_place['orderContainers'][0]['prices']:
            cost_key = 'hourlyRecurringFee' if billing == 'hourly' else 'recurringFee'
            table.add_row([
                price['item']['keyName'],
                price['item']['description'],
                price[cost_key] if cost_key in price else formatting.blank()
            ])

    else:
        if not (env.skip_confirmations or formatting.confirm(
                "This action will incur charges on your account. Continue?")):
            raise exceptions.CLIAbort("Aborting order.")

        order = manager.place_order(*args, **kwargs)

        table = formatting.KeyValueTable(['name', 'value'])
        table.align['name'] = 'r'
        table.align['value'] = 'l'
        table.add_row(['id', order['orderId']])
        table.add_row(['created', order['orderDate']])
        table.add_row(['status', order['placedOrder']['status']])
    env.fout(table)



"""
    SoftLayer.CLI.custom_types
    ~~~~~~~~~~~~~~~~~~~~~~~~
    Custom type declarations extending click.ParamType

    :license: MIT, see LICENSE for more details.
"""

import click


# pylint: disable=inconsistent-return-statements
class NetworkParamType(click.ParamType):
    """Validates a network parameter type and converts to a tuple.

    todo: Implement to ipaddress.ip_network once the ipaddress backport
          module can be added as a dependency or is available on all
          supported python versions.
    """
    name = 'network'

    def convert(self, value, param, ctx):
        try:
            # Inlined from python standard ipaddress module
            # https://docs.python.org/3/library/ipaddress.html
            address = str(value).split('/')
            if len(address) != 2:
                raise ValueError("Only one '/' permitted in %r" % value)

            ip_address, cidr = address
            return (ip_address, int(cidr))
        except ValueError:
            self.fail(f'{value} is not a valid network', param, ctx)



"""Delete the quote of an order"""
# :license: MIT, see LICENSE for more details.
import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.managers import ordering


@click.command(cls=SLCommand)
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Delete the quote of an order"""

    manager = ordering.OrderingManager(env.client)
    result = manager.delete_quote(identifier)

    if result:
        env.fout(f"Quote id: {identifier} was deleted.")



"""Add a server to the firewall"""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('target')
@click.option('--firewall-type', type=click.Choice(['vs', 'vlan', 'server']), help='Firewall type', required=True)
@click.option("-h", '--high-availability', is_flag=True, help='High available firewall option')
@click.option('-f', '--force',  default=False, is_flag=True, help="Force addition of firewall to the server")
@environment.pass_env
def cli(env, target, firewall_type, high_availability, force):
    """Create new firewall.

    TARGET: Id of the server the firewall will protect
    """

    mgr = SoftLayer.FirewallManager(env.client)
    pkg = {}
    if not env.skip_confirmations:
        if firewall_type == 'vlan':
            pkg = mgr.get_dedicated_package(ha_enabled=high_availability)
        elif firewall_type == 'vs':
            pkg = mgr.get_standard_package(target, is_virt=True)
        elif firewall_type == 'server':
            pkg = mgr.get_standard_package(target, is_virt=False)

        if not pkg:
            exceptions.CLIAbort("Unable to add firewall - Is network public enabled?")

        click.echo("******************")
        click.echo("Product: %s" % pkg[0]['description'])
        click.echo("Price: $%s monthly" % pkg[0]['prices'][0]['recurringFee'])
        click.echo("******************")

        if not force:
            if not formatting.confirm("This action will incur charges on your account. Continue?"):
                raise exceptions.CLIAbort('Aborted.')

    if firewall_type == 'vlan':
        mgr.add_vlan_firewall(target, ha_enabled=high_availability)
    elif firewall_type == 'vs':
        mgr.add_standard_firewall(target, is_virt=True)
    elif firewall_type == 'server':
        mgr.add_standard_firewall(target, is_virt=False)

    click.echo("Firewall is being created!")



"""Get event logs relating to security groups"""
# :license: MIT, see LICENSE for more details.

import json

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting

COLUMNS = ['event', 'label', 'date', 'metadata']


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('request_id')
@environment.pass_env
def get_by_request_id(env, request_id):
    """Search for event logs by request id"""
    mgr = SoftLayer.NetworkManager(env.client)

    logs = mgr.get_event_logs_by_request_id(request_id)

    table = formatting.Table(COLUMNS)
    table.align['metadata'] = "l"

    for log in logs:
        metadata = json.dumps(json.loads(log['metaData']), indent=4, sort_keys=True)

        table.add_row([log['eventName'], log['label'], log['eventCreateDate'], metadata])

    env.fout(table)



"""List Users."""
# :license: MIT, see LICENSE for more details.


import json

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('user')
@click.option('--template', '-t', required=True,
              help="A json string describing https://softlayer.github.io/reference/datatypes/SoftLayer_User_Customer/")
@environment.pass_env
def cli(env, user, template):
    """Edit a Users details

    JSON strings should be enclosed in '' and each item should be enclosed in ""

    Example::

        slcli user edit-details testUser -t '{"firstName": "Test", "lastName": "Testerson"}'

    """
    mgr = SoftLayer.UserManager(env.client)
    user_id = helpers.resolve_id(mgr.resolve_ids, user, 'username')

    user_template = {}
    if template is not None:
        try:
            template_object = json.loads(template)
            for key in template_object:
                user_template[key] = template_object[key]
        except ValueError as ex:
            raise exceptions.ArgumentError(f"Unable to parse --template. {ex}")

    result = mgr.edit_user(user_id, user_template)
    if result:
        click.secho(f"{user} updated successfully", fg='green')
    else:
        click.secho(f"Failed to update {user}", fg='red')



"""
    SoftLayer.transports.debug
    ~~~~~~~~~~~~~~~~~~~~
    Debugging transport. Will print out verbose logging information.

    :license: MIT, see LICENSE for more details.
"""

import json
import logging
import time

from SoftLayer import exceptions


class DebugTransport(object):
    """Transport that records API call timings."""

    def __init__(self, transport):
        self.transport = transport

        #: List All API calls made during a session
        self.requests = []
        self.logger = logging.getLogger(__name__)

    def __call__(self, call):
        call.start_time = time.time()

        self.pre_transport_log(call)
        try:
            call.result = self.transport(call)
        except (exceptions.SoftLayerAPIError, exceptions.TransportError) as ex:
            call.exception = ex

        self.post_transport_log(call)

        call.end_time = time.time()
        self.requests.append(call)

        if call.exception is not None:
            self.logger.debug(self.print_reproduceable(call))
            raise call.exception

        return call.result

    def pre_transport_log(self, call):
        """Prints a warning before calling the API """
        output = f"Calling: {call}"
        self.logger.warning(output)

    def post_transport_log(self, call):
        """Prints the result "Returned Data: \n%s" % (call.result)of an API call"""
        output = f"Returned Data: \n{json.dumps(call.result)}"
        self.logger.debug(output)

    def get_last_calls(self):
        """Returns all API calls for a session"""
        return self.requests

    def print_reproduceable(self, call):
        """Prints a reproduceable debugging output"""
        return self.transport.print_reproduceable(call)



"""Failover to a replicant volume."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('volume-id')
@click.option('--replicant-id', help="ID of the replicant volume.")
@environment.pass_env
def cli(env, volume_id, replicant_id):
    """Failover a block volume to the given replica volume."""
    block_storage_manager = SoftLayer.BlockStorageManager(env.client)

    success = block_storage_manager.failover_to_replicant(
        volume_id,
        replicant_id
    )

    if success:
        click.echo("Failover to replicant is now in progress.")
    else:
        click.echo("Failover operation could not be initiated.")



"""Get details for a hardware device."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--start_date', '-s', type=click.STRING, required=True,
              help="Start Date YYYY-MM-DD, YYYY-MM-DDTHH:mm:ss,")
@click.option('--end_date', '-e', type=click.STRING, required=True,
              help="End Date YYYY-MM-DD, YYYY-MM-DDTHH:mm:ss")
@click.option('--summary_period', '-p', type=click.INT, default=3600, show_default=True,
              help="300, 600, 1800, 3600, 43200 or 86400 seconds")
@click.option('--quite_summary', '-q', is_flag=True, default=False, show_default=True,
              help="Only show the summary table")
@environment.pass_env
def cli(env, identifier, start_date, end_date, summary_period, quite_summary):
    """Bandwidth data over date range. Bandwidth is listed in GB

    Using just a date might get you times off by 1 hour, use T00:01 to get just the specific days data
    Timezones can also be included with the YYYY-MM-DDTHH:mm:ss.00000-HH:mm format.

    Due to some rounding and date alignment details, results here might be slightly different than
    results in the control portal.

    Example::

        slcli vs bandwidth 1234 -s 2019-05-01T00:01 -e 2019-05-02T00:00:01.00000-12:00
    """
    vsi = SoftLayer.VSManager(env.client)
    vsi_id = helpers.resolve_id(vsi.resolve_ids, identifier, 'VS')

    data = vsi.get_bandwidth_data(vsi_id, start_date, end_date, None, None)

    title = f"Bandwidth Report: {start_date} - {end_date}"
    table, sum_table = create_bandwidth_table(data, summary_period, title)

    env.fout(sum_table)
    if not quite_summary:
        env.fout(table)


def create_bandwidth_table(data, summary_period, title="Bandwidth Report"):
    """Create 2 tables, bandwidth and sumamry. Used here and in hw bandwidth command"""

    formatted_data = {}
    for point in data:
        key = utils.clean_time(point['dateTime'])
        data_type = point['type']
        # conversion from byte to megabyte
        value = round(float(point['counter']) / 2 ** 20, 4)
        if formatted_data.get(key) is None:
            formatted_data[key] = {}
        formatted_data[key][data_type] = float(value)

    table = formatting.Table(['Date', 'Pub In', 'Pub Out', 'Pri In', 'Pri Out'], title=title)

    sum_table = formatting.Table(['Type', 'Sum GB', 'Average MBps', 'Max GB', 'Max Date'], title="Summary")

    # Required to specify keyName because getBandwidthTotals returns other counter types for some reason.
    bw_totals = [
        {'keyName': 'publicIn_net_octet', 'sum': 0.0, 'max': 0, 'name': 'Pub In'},
        {'keyName': 'publicOut_net_octet', 'sum': 0.0, 'max': 0, 'name': 'Pub Out'},
        {'keyName': 'privateIn_net_octet', 'sum': 0.0, 'max': 0, 'name': 'Pri In'},
        {'keyName': 'privateOut_net_octet', 'sum': 0.0, 'max': 0, 'name': 'Pri Out'},
    ]

    for key, value in formatted_data.items():
        new_row = [key]
        for bw_type in bw_totals:
            counter = value.get(bw_type['keyName'], 0)
            new_row.append(mb_to_gb(counter))
            bw_type['sum'] = bw_type['sum'] + counter
            if counter > bw_type['max']:
                bw_type['max'] = counter
                bw_type['maxDate'] = key
        table.add_row(new_row)

    for bw_type in bw_totals:
        total = bw_type.get('sum', 0.0)
        average = 0
        if total > 0:
            average = round(total / len(formatted_data) / summary_period, 4)
        sum_table.add_row([
            bw_type.get('name'),
            mb_to_gb(total),
            average,
            mb_to_gb(bw_type.get('max')),
            bw_type.get('maxDate')
        ])

    return table, sum_table


def mb_to_gb(mbytes):
    """Converts a MegaByte int to GigaByte. mbytes/2^10"""
    return round(mbytes / 2 ** 10, 4)



"""
    SoftLayer.block
    ~~~~~~~~~~~~~~~
    Block Storage Manager

    :license: MIT, see LICENSE for more details.
"""
from SoftLayer.managers.storage import StorageManager
from SoftLayer.managers import storage_utils
from SoftLayer import utils


# pylint: disable=too-many-public-methods


class BlockStorageManager(StorageManager):
    """Manages SoftLayer Block Storage volumes.

    See product information here: https://www.ibm.com/cloud/block-storage
    """

    def list_block_volume_limit(self):
        """Returns a list of block volume count limit.

        :return: Returns a list of block volume count limit.
        """
        return self.get_volume_count_limits()

    def list_block_volumes(self, datacenter=None, username=None, storage_type=None, order=None, **kwargs):
        """Returns a list of block volumes.

        :param datacenter: Datacenter short name (e.g.: dal09)
        :param username: Name of volume.
        :param storage_type: Type of volume: Endurance or Performance
        :param order: Volume order id.
        :param kwargs:
        :return: Returns a list of block volumes.
        """
        if 'mask' not in kwargs:
            items = [
                'id',
                'username',
                'lunId',
                'capacityGb',
                'bytesUsed',
                'serviceResource.datacenter[name]',
                'serviceResourceBackendIpAddress',
                'activeTransactionCount',
                'replicationPartnerCount'
            ]
            kwargs['mask'] = ','.join(items)

        _filter = utils.NestedDict(kwargs.get('filter') or {})

        _filter['iscsiNetworkStorage']['serviceResource']['type']['type'] = utils.query_filter('!~ ISCSI')
        _filter['iscsiNetworkStorage']['id'] = utils.query_filter_orderby()

        _filter['iscsiNetworkStorage']['storageType']['keyName'] = utils.query_filter('*BLOCK_STORAGE*')
        if storage_type:
            _filter['iscsiNetworkStorage']['storageType']['keyName'] = (
                utils.query_filter('%s_BLOCK_STORAGE*' % storage_type.upper()))

        if datacenter:
            _filter['iscsiNetworkStorage']['serviceResource']['datacenter']['name'] = utils.query_filter(datacenter)

        if username:
            _filter['iscsiNetworkStorage']['username'] = utils.query_filter(username)

        if order:
            _filter['iscsiNetworkStorage']['billingItem']['orderItem']['order']['id'] = utils.query_filter(order)

        kwargs['filter'] = _filter.to_dict()
        return self.client.call('Account', 'getIscsiNetworkStorage', iter=True, **kwargs)

    def get_block_volume_details(self, volume_id, **kwargs):
        """Returns details about the specified volume.

        :param volume_id: ID of volume.
        :param kwargs:
        :return: Returns details about the specified volume.
        """
        return self.get_volume_details(volume_id, **kwargs)

    def get_block_volume_access_list(self, volume_id, **kwargs):
        """Returns a list of authorized hosts for a specified volume.

        :param volume_id: ID of volume.
        :param kwargs:
        :return: Returns a list of authorized hosts for a specified volume.
        """
        return self.get_volume_access_list(volume_id, **kwargs)

    def get_block_volume_snapshot_list(self, volume_id, **kwargs):
        """Returns a list of snapshots for the specified volume.

        :param volume_id: ID of volume.
        :param kwargs:
        :return: Returns a list of snapshots for the specified volume.
        """
        return self.get_volume_snapshot_list(volume_id, **kwargs)

    def assign_subnets_to_acl(self, access_id, subnet_ids):
        """Assigns subnet records to ACL for the access host.

        access_id is the host_id obtained by: slcli block access-list <volume_id>

        :param integer access_id: id of the access host
        :param list subnet_ids: The ids of the subnets to be assigned
        :return: Returns int array of assigned subnet ids
        """
        return self.client.call('Network_Storage_Allowed_Host', 'assignSubnetsToAcl', subnet_ids, id=access_id)

    def remove_subnets_from_acl(self, access_id, subnet_ids):
        """Removes subnet records from ACL for the access host.

        access_id is the host_id obtained by: slcli block access-list <volume_id>

        :param integer access_id: id of the access host
        :param list subnet_ids: The ids of the subnets to be removed
        :return: Returns int array of removed subnet ids
        """
        return self.client.call('Network_Storage_Allowed_Host', 'removeSubnetsFromAcl', subnet_ids, id=access_id)

    def get_subnets_in_acl(self, access_id):
        """Returns a list of subnet records for the access host.

         access_id is the host_id obtained by: slcli block access-list <volume_id>

        :param integer access_id: id of the access host
        :return: Returns an array of SoftLayer_Network_Subnet objects
        """
        return self.client.call('Network_Storage_Allowed_Host', 'getSubnetsInAcl', id=access_id)

    def order_block_volume(self, storage_type, location, size, os_type,
                           iops=None, tier_level=None, snapshot_size=None,
                           service_offering='storage_as_a_service',
                           hourly_billing_flag=False):
        """Places an order for a block volume.

        :param storage_type: 'performance' or 'endurance'
        :param location: Datacenter in which to order iSCSI volume
        :param size: Size of the desired volume, in GB
        :param os_type: OS Type to use for volume alignment, see help for list
        :param iops: Number of IOPs for a "Performance" order
        :param tier_level: Tier level to use for an "Endurance" order
        :param snapshot_size: The size of optional snapshot space,
            if snapshot space should also be ordered (None if not ordered)
        :param service_offering: Requested offering package to use in the order
            ('storage_as_a_service', 'enterprise', or 'performance')
        :param hourly_billing_flag: Billing type, monthly (False)
            or hourly (True), default to monthly.
        """
        order = storage_utils.prepare_volume_order_object(
            self, storage_type, location, size, iops, tier_level,
            snapshot_size, service_offering, 'block', hourly_billing_flag
        )

        order['osFormatType'] = {'keyName': os_type}

        return self.client.call('Product_Order', 'placeOrder', order)

    def cancel_block_volume(self, volume_id, reason='No longer needed', immediate=False):
        """Cancels the given block storage volume.

        :param integer volume_id: The volume ID
        :param string reason: The reason for cancellation
        :param boolean immediate_flag: Cancel immediately or on anniversary date
        """
        return self.cancel_volume(volume_id, reason, immediate)

    def set_credential_password(self, access_id, password):
        """Sets the password for an access host

        :param integer access_id: id of the access host
        :param string password: password to  set
        """

        return self.client.call('Network_Storage_Allowed_Host', 'setCredentialPassword',
                                password, id=access_id)

    def create_or_update_lun_id(self, volume_id, lun_id):
        """Set the LUN ID on a volume.

        :param integer volume_id: The id of the volume
        :param integer lun_id: LUN ID to set on the volume
        :return: a SoftLayer_Network_Storage_Property object
        """
        return self.client.call('Network_Storage', 'createOrUpdateLunId', lun_id, id=volume_id)

    def _get_ids_from_username(self, username):
        object_mask = "mask[id]"
        results = self.list_block_volumes(username=username, mask=object_mask)
        if results:
            return [result['id'] for result in results]
        return []

    def get_cloud_list(self):
        """Returns a list cloud object storage.

        return: Returns a list cloud object storage.
        """

        mask = 'mask[id,username,billingItem,storageType, notes]'

        return self.client.call('Account', 'getHubNetworkStorage', mask=mask)

    def get_buckets(self, object_id):
        """Return  buckets data of the cloud storage.

        :param object_id cloud object storage identifier

        Returns: Get buckets

        """
        return self.client.call('SoftLayer_Network_Storage_Hub_Cleversafe_Account', 'getBuckets', id=object_id)



"""Create a support ticket."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers
from SoftLayer.CLI import ticket


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--title', required=True, help="The title of the ticket")
@click.option('--subject-id', type=int, required=True,
              help="""The subject id to use for the ticket, run 'slcli ticket subjects' to get the list""")
@click.option('--body', help="The ticket body")
@click.option('--hardware', 'hardware_identifier',
              help="The identifier for hardware to attach")
@click.option('--virtual', 'virtual_identifier',
              help="The identifier for a virtual server to attach")
@click.option('--priority', 'priority', type=click.Choice(['1', '2', '3', '4']), default=None,
              help="""Ticket priority, from 1 (Critical) to 4 (Minimal Impact).
              Only settable with Advanced and Premium support. See https://www.ibm.com/cloud/support""")
@environment.pass_env
def cli(env, title, subject_id, body, hardware_identifier, virtual_identifier, priority):
    """Create a Infrastructure support ticket.

    Will create the ticket with `Some text`.::

        slcli ticket create --body="Some text" --subject-id 1522 --hardware 12345 --title "My New Ticket"

    Will create the ticket with text from STDIN::

        cat sometfile.txt | slcli ticket create --subject-id 1003 --virtual 111111 --title "Reboot Me"

    Will open the default text editor, and once closed, use that text to create the ticket::

        slcli ticket create --subject-id 1482 --title "Vyatta Questions..."

    """
    ticket_mgr = SoftLayer.TicketManager(env.client)
    if body is None:
        stdin = click.get_text_stream('stdin')
        # Means there is text on the STDIN buffer, read it and add to the ticket
        if not stdin.isatty():
            body = stdin.read()
        # This is an interactive terminal, open a text editor
        else:
            body = click.edit('\n\n' + ticket.TEMPLATE_MSG)
    created_ticket = ticket_mgr.create_ticket(
        title=title,
        body=body,
        subject=subject_id,
        priority=priority)

    if hardware_identifier:
        hardware_mgr = SoftLayer.HardwareManager(env.client)
        hardware_id = helpers.resolve_id(hardware_mgr.resolve_ids, hardware_identifier, 'hardware')
        ticket_mgr.attach_hardware(created_ticket['id'], hardware_id)

    if virtual_identifier:
        vs_mgr = SoftLayer.VSManager(env.client)
        vs_id = helpers.resolve_id(vs_mgr.resolve_ids, virtual_identifier, 'VS')
        ticket_mgr.attach_virtual_server(created_ticket['id'], vs_id)

    env.fout(ticket.get_ticket_results(ticket_mgr, False, created_ticket['id']))



createCancelServerTicket = {'id': 1234, 'title': 'Server Cancellation Request'}
getObject = {
    "guestNetworkComponentId": 1705294,
    "id": 1234,
    "status": "allow_edit",
    "billingItem": {
        "allowCancellationFlag": 1,
        "associatedBillingItemId": "20952512",
        "categoryCode": "firewall",
        "createDate": "2014-03-21T14:07:04-05:00",
        "cycleStartDate": "2014-03-21T14:07:04-05:00",
        "description": "10Mbps Hardware Firewall",
        "id": 21370814,
        "laborFee": "0",
        "laborFeeTaxRate": ".066",
        "lastBillDate": "2014-03-21T14:07:04-05:00",
        "modifyDate": "2014-03-21T14:07:07-05:00",
        "nextBillDate": "2014-04-04T00:00:00-05:00",
        "oneTimeFee": "0",
        "oneTimeFeeTaxRate": ".066",
        "orderItemId": 28712824,
        "parentId": 20952512,
        "recurringFee": "0",
        "recurringFeeTaxRate": ".066",
        "recurringMonths": 1,
        "serviceProviderId": 1,
        "setupFee": "0",
        "setupFeeTaxRate": ".066"
    },
    "guestNetworkComponent": {
        "createDate": "2014-03-17T13:49:00-05:00",
        "guestId": 3895386,
        "id": 1705294,
        "macAddress": "06:a4:8d:d2:88:34",
        "maxSpeed": 10,
        "modifyDate": "2014-03-17T13:49:20-05:00",
        "name": "eth",
        "networkId": 1310218,
        "port": 1,
        "speed": 10,
        "status": "ACTIVE",
        "uuid": "3f1b5e08-a652-fb3b-1baa-8ace70c90fe9",
        "guest": {
            "accountId": 307608,
            "dedicatedAccountHostOnlyFlag": False,
            "domain": "test.com",
            "fullyQualifiedDomainName": "test.test.com",
            "hostname": "firewalltest",
            "id": 3895386,
            "maxCpu": 1,
            "maxCpuUnits": "CORE",
            "maxMemory": 1024,
            "modifyDate": "2014-03-21T14:05:51-05:00",
            "startCpus": 1,
            "statusId": 1001,
            "uuid": "29b40ef0-a43a-8cb6-31be-1878cb6853f0",
            "status": {
                "keyName": "ACTIVE",
                "name": "Active"
            }
        }
    }
}

getRules = [
    {
        'destinationIpAddress': 'any on server',
        'protocol': 'tcp',
        'orderValue': 1,
        'destinationIpSubnetMask': '255.255.255.255',
        'destinationPortRangeStart': 80,
        'sourceIpSubnetMask': '0.0.0.0',
        'destinationPortRangeEnd': 80,
        'version': 4,
        'action': 'permit',
        'sourceIpAddress': '0.0.0.0'
    },
    {
        'destinationIpAddress': 'any on server',
        'protocol': 'tcp',
        'orderValue': 2,
        'destinationIpSubnetMask': '255.255.255.255',
        'destinationPortRangeStart': 1,
        'sourceIpSubnetMask': '255.255.255.255',
        'destinationPortRangeEnd': 65535,
        'version': 4,
        'action': 'permit',
        'sourceIpAddress': '193.212.1.10'
    },
    {
        'destinationIpAddress': 'any on server',
        'protocol': 'tcp',
        'orderValue': 3,
        'destinationIpSubnetMask': '255.255.255.255',
        'destinationPortRangeStart': 80,
        'sourceIpSubnetMask': '0.0.0.0',
        'destinationPortRangeEnd': 800,
        'version': 4,
        'action': 'permit',
        'sourceIpAddress': '0.0.0.0'
    }
]

getBillingItem = {"id": 21370814}



"""Delete an image."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Delete an image."""

    image_mgr = SoftLayer.ImageManager(env.client)
    image_id = helpers.resolve_id(image_mgr.resolve_ids, identifier, 'image')

    image_mgr.delete_image(image_id)



"""
    SoftLayer.CLI.shell.core
    ~~~~~~~~~~~~~~~~~~~~~~~~
    An interactive shell which exposes the CLI

    :license: MIT, see LICENSE for more details.
"""
import copy
import os
import shlex
import sys
import traceback

import click
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit import PromptSession

from SoftLayer.CLI import core
from SoftLayer.CLI import environment
from SoftLayer.shell import completer
from SoftLayer.shell import routes

# pylint: disable=broad-except


class ShellExit(Exception):
    """Exception raised to quit the shell."""


@click.command()
@environment.pass_env
@click.pass_context
def cli(ctx, env):
    """Enters a shell for slcli."""

    # Set up the environment
    env = copy.deepcopy(env)
    env.load_modules_from_python(routes.ALL_ROUTES)
    env.aliases.update(routes.ALL_ALIASES)
    env.vars['global_args'] = ctx.parent.params
    env.vars['is_shell'] = True
    env.vars['last_exit_code'] = 0

    # Set up prompt_toolkit settings
    app_path = click.get_app_dir('softlayer_shell')
    if not os.path.exists(app_path):
        os.makedirs(app_path)
    complete = completer.ShellCompleter(core.cli)

    session = PromptSession()

    while True:
        try:
            line = session.prompt(
                completer=complete,
                complete_while_typing=True,
                auto_suggest=AutoSuggestFromHistory(),
            )

            # Parse arguments
            try:
                args = shlex.split(line)
            except ValueError as ex:
                print("Invalid Command: %s" % ex)
                continue

            if not args:
                continue

            # Run Command
            try:
                # Reset client so that the client gets refreshed
                env.client = None
                core.main(args=list(get_env_args(env)) + args,
                          obj=env,
                          prog_name="",
                          reraise_exceptions=True)
            except SystemExit as ex:
                env.vars['last_exit_code'] = ex.code
            except EOFError:
                return
            except ShellExit:
                return
            except Exception:
                env.vars['last_exit_code'] = 1
                traceback.print_exc(file=sys.stderr)

        except KeyboardInterrupt:
            env.vars['last_exit_code'] = 130


def get_env_args(env):
    """Yield options to inject into the slcli command from the environment."""
    for arg, val in env.vars.get('global_args', {}).items():
        if val is True:
            yield '--%s' % arg
        elif isinstance(val, int):
            for _ in range(val):
                yield '--%s' % arg
        elif val is None:
            continue
        else:
            yield '--%s=%s' % (arg, val)



"""Details of a Tag."""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI.tags.list import detailed_table
from SoftLayer.managers.tags import TagManager


@click.command(cls=SLCommand)
@click.argument('identifier')
@click.option('--name', required=False, default=False, is_flag=True, show_default=False,
              help='Assume identifier is a tag name. Useful if your tag name is a number.')
@environment.pass_env
def cli(env, identifier, name):
    """Get details for a Tag. Identifier can be either a name or tag-id"""

    tag_manager = TagManager(env.client)

    # If the identifier is a int, and user didn't tell us it was a name.
    if str.isdigit(identifier) and not name:
        tags = [tag_manager.get_tag(identifier)]
    else:
        tags = tag_manager.get_tag_by_name(identifier)
    table = detailed_table(tag_manager, tags)
    env.fout(table)



"""Lists Email Delivery Service """
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.account import AccountManager
from SoftLayer.managers.email import EmailManager
from SoftLayer import utils


@click.command(cls=SLCommand)
@environment.pass_env
def cli(env):
    """Lists Email Delivery Service """
    manager = AccountManager(env.client)
    email_manager = EmailManager(env.client)
    result = manager.get_network_message_delivery_accounts()

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'
    table_information = formatting.KeyValueTable(['id', 'username', 'hostname', 'description', 'vendor'])
    table_information.align['id'] = 'r'
    table_information.align['username'] = 'l'

    for email in result:
        table_information.add_row([email.get('id'), email.get('username'), email.get('emailAddress'),
                                   utils.lookup(email, 'type', 'description'),
                                   utils.lookup(email, 'vendor', 'keyName')])

        overview_table = _build_overview_table(email_manager.get_account_overview(email.get('id')))
        # Commented this line until we fix EmailManager.GetStatistics() method in golang plugin
        # statistics = email_manager.get_statistics(email.get('id'))

        table.add_row(['email_information', table_information])
        table.add_row(['email_overview', overview_table])
        # Commented these lines until we fix EmailManager.GetStatistics() method in golang plugin
        # for statistic in statistics:
        #     table.add_row(['statistics', build_statistics_table(statistic)])

    env.fout(table)


def _build_overview_table(email_overview):
    table = formatting.Table(
        ['package', 'reputation'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    table.add_row([
        utils.lookup(email_overview, 'profile', 'package'),
        utils.lookup(email_overview, 'profile', 'reputation')
    ])

    return table


def build_statistics_table(statistics):
    """statistics records of Email Delivery account"""
    table = formatting.Table(['delivered', 'requests', 'bounces', 'opens', 'clicks', 'spam_reports'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    table.add_row([statistics.get('delivered'), statistics.get('requests'),
                   statistics.get('bounces'), statistics.get('opens'),
                   statistics.get('clicks'), statistics.get('spamReports')])

    return table



"""List SSL certificates."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--status',
              default="all",
              show_default=True,
              type=click.Choice(['all', 'valid', 'expired']),
              help="Show certificates with this status")
@click.option('--sortby',
              type=click.Choice(['id',
                                 'common_name',
                                 'days_until_expire',
                                 'notes']),
              help="Column to sort by")
@environment.pass_env
def cli(env, status, sortby):
    """List SSL certificates."""
    manager = SoftLayer.SSLManager(env.client)

    certificates = manager.list_certs(status)

    table = formatting.Table(['id',
                              'common_name',
                              'days_until_expire',
                              'notes'])
    for certificate in certificates:
        table.add_row([
            certificate['id'],
            certificate['commonName'],
            certificate['validityDays'],
            certificate.get('notes', formatting.blank())
        ])
    table.sortby = sortby
    env.fout(table)



"""Check if a virtual server is ready."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--wait', default=0, show_default=True, type=click.INT, help="Seconds to wait")
@environment.pass_env
def cli(env, identifier, wait):
    """Check if a virtual server is ready."""

    vsi = SoftLayer.VSManager(env.client)
    vs_id = helpers.resolve_id(vsi.resolve_ids, identifier, 'VS')
    ready = vsi.wait_for_ready(vs_id, wait)
    if ready:
        env.fout("READY")
    else:
        raise exceptions.CLIAbort(f"Instance {vs_id} not ready")



"""Create a Reserved Capacity instance."""

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.vs_capacity import CapacityManager as CapacityManager


@click.command(cls=SLCommand)
@click.option('--name', '-n', required=True, prompt=True,
              help="Name for your new reserved capacity")
@click.option('--backend_router_id', '-b', required=True, prompt=True, type=int,
              help="backendRouterId, create-options has a list of valid ids to use.")
@click.option('--flavor', '-f', required=True, prompt=True,
              help="Capacity keyname (C1_2X2_1_YEAR_TERM for example).")
@click.option('--instances', '-i', required=True, prompt=True, type=int,
              help="Number of VSI instances this capacity reservation can support.")
@click.option('--test', is_flag=True,
              help="Do not actually create the virtual server")
@environment.pass_env
def cli(env, name, backend_router_id, flavor, instances, test=False):
    """Create a Reserved Capacity instance.

    [red underline]*WARNING*[/]
    [red]Reserved Capacity is on a yearly contract and not cancelable until the contract is expired.[/]
    """
    manager = CapacityManager(env.client)

    result = manager.create(
        name=name,
        backend_router_id=backend_router_id,
        flavor=flavor,
        instances=instances,
        test=test)
    if test:
        table = formatting.Table(['Name', 'Value'], "Test Order")
        container = result['orderContainers'][0]
        table.add_row(['Name', container['name']])
        table.add_row(['Location', container['locationObject']['longName']])
        for price in container['prices']:
            table.add_row(['Contract', price['item']['description']])
        table.add_row(['Hourly Total', result['postTaxRecurring']])
    else:
        table = formatting.Table(['Name', 'Value'], "Reciept")
        table.add_row(['Order Date', result['orderDate']])
        table.add_row(['Order ID', result['orderId']])
        table.add_row(['status', result['placedOrder']['status']])
        table.add_row(['Hourly Total', result['orderDetails']['postTaxRecurring']])
    env.fout(table)



"""Order a file storage replica volume."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers
from SoftLayer import utils


CONTEXT_SETTINGS = {'token_normalize_func': lambda x: x.upper()}


@click.command(cls=SoftLayer.CLI.command.SLCommand, context_settings=CONTEXT_SETTINGS)
@click.argument('volume_id')
@click.option('--snapshot-schedule', '-s',
              help='Snapshot schedule to use for replication, '
              '(HOURLY | DAILY | WEEKLY)',
              required=True,
              type=click.Choice(['HOURLY', 'DAILY', 'WEEKLY']))
@click.option('--location', '-l',
              help='Short name of the data center for the replicant '
              '(e.g.: dal09)',
              required=True)
@click.option('--tier',
              help='Endurance Storage Tier (IOPS per GB) of the primary'
              ' volume for which a replicant is ordered [optional]',
              type=click.Choice(['0.25', '2', '4', '10']))
@click.option('--iops', '-i', type=int,
              help='Performance Storage IOPs, between 100 and 6000 in multiples of 100,'
              'if no IOPS value is specified, the IOPS value of the original volume will be used',
              )
@click.option('--force', default=False, is_flag=True, help="Force cancel block volume without confirmation")
@environment.pass_env
def cli(env, volume_id, snapshot_schedule, location, tier, iops, force):
    """Order a file storage replica volume.

    Example::
        slcli file replica-order 12345678 -s DAILY -d dal09 --tier 4 --os-type LINUX
        This command orders a replica for volume with ID 12345678, which performs DAILY
        replication, is located at dal09, tier level is 4, OS type is Linux.
    """
    file_manager = SoftLayer.FileStorageManager(env.client)
    file_volume_id = helpers.resolve_id(file_manager.resolve_ids, volume_id, 'File Storage')

    if tier is not None:
        tier = float(tier)

    if iops is not None:
        if iops < 100 or iops > 6000:
            raise exceptions.CLIHalt("-i|--iops must be between 100 and 6000, inclusive. "
                                     "Run 'slcli block volume-options' to check available options.")

        if iops % 100 != 0:
            raise exceptions.CLIHalt("-i|--iops must be a multiple of 100. "
                                     "Run 'slcli block volume-options' to check available options.")

    if not force:
        if not (env.skip_confirmations or
                formatting.confirm(f"This will Order a file storage replica volume: {volume_id} "
                                   "and cannot be undone. Continue?")):
            raise exceptions.CLIAbort('Aborted.')

    try:
        order = file_manager.order_replicant_volume(
            file_volume_id,
            snapshot_schedule=snapshot_schedule,
            location=location,
            tier=tier, iops=iops,
        )
    except ValueError as ex:
        raise exceptions.ArgumentError(str(ex))

    if 'placedOrder' in order.keys():
        click.echo(f"Order #{utils.lookup(order, 'placedOrder', 'id')} placed successfully!")
        for item in utils.lookup(order, 'placedOrder', 'items'):
            click.echo(" > %s" % item.get('description'))
    else:
        click.echo("Order could not be placed! Please verify your options " +
                   "and try again.")



"""Removes a user access to a given device."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--hardware', help="Hardware ID")
@click.option('--virtual', help="Virtual Guest ID")
@click.option('--dedicated', help="Dedicated host ID ")
@environment.pass_env
def cli(env, identifier, hardware, virtual, dedicated):
    """Removes a user access to a given device.

    Example: slcli user remove-access 123456 --hardware 123456789
    """

    mgr = SoftLayer.UserManager(env.client)
    result = False
    if hardware:
        result = mgr.remove_hardware_access(identifier, hardware)
        if result:
            click.secho(f"Removed access to hardware: {hardware}.", fg='green')

    if virtual:
        result = mgr.remove_virtual_access(identifier, virtual)
        if result:
            click.secho(f"Removed access to virtual guest: {virtual}", fg='green')

    if dedicated:
        result = mgr.remove_dedicated_access(identifier, dedicated)
        if result:
            click.secho(f"Removed access to dedicated host: {dedicated}", fg='green')

    if not result:
        raise exceptions.CLIAbort('A device option is required.\n'
                                  'E.g slcli user remove-access 123456 --hardware 91803794\n'
                                  '    slcli user remove-access 123456 --dedicated 91803793\n'
                                  '    slcli user remove-access 123456 --virtual 91803792')



"""Delete load balancer protocol."""
import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--uuid', help="Load Balancer Uuid.")
@environment.pass_env
def cli(env, identifier, uuid):
    """delete a load balancer protocol."""

    mgr = SoftLayer.LoadBalancerManager(env.client)

    uuid_lb = mgr.get_lb(identifier)['uuid']

    protocol = mgr.delete_protocol(uuid_lb, uuid)

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'
    table.add_row(['Id', protocol.get('id')])
    table.add_row(['UUI', protocol.get('uuid')])
    table.add_row(['Address', protocol.get('address')])
    table.add_row(['Type', SoftLayer.LoadBalancerManager.TYPE.get(protocol.get('type'))])
    table.add_row(['Description', protocol.get('description')])

    env.fout(table)



"""Place quote"""
# :license: MIT, see LICENSE for more details.

import json

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.managers import ordering


@click.command(cls=SLCommand)
@click.argument('package_keyname')
@click.argument('location')
@click.option('--preset',
              help="The order preset (if required by the package)")
@click.option('--name',
              help="A custom name to be assigned to the quote (optional)")
@click.option('--send-email',
              is_flag=True,
              help="The quote will be sent to the email address associated with your user.")
@click.option('--complex-type',
              help="The complex type of the order. Starts with 'SoftLayer_Container_Product_Order'.")
@click.option('--extras',
              help="JSON string denoting extra data that needs to be sent with the order")
@click.argument('order_items', nargs=-1)
@environment.pass_env
def cli(env, package_keyname, location, preset, name, send_email, complex_type,
        extras, order_items):
    """Place a quote.

    This CLI command is used for creating  a quote of the specified package in
    the given location (denoted by a datacenter's long name). Orders made via the CLI
    can then be converted to be made programmatically by calling
    SoftLayer.OrderingManager.place_quote() with the same keynames.

    Packages for ordering can be retrieved from `slcli order package-list`
    Presets for ordering can be retrieved from `slcli order preset-list` (not all packages
    have presets)

    Items can be retrieved from `slcli order item-list`. In order to find required
    items for the order, use `slcli order category-list`, and then provide the
    --category option for each category code in `slcli order item-list`.


    Example::

        # Place quote a VSI with 4 CPU, 16 GB RAM, 100 GB SAN disk,
        # Ubuntu 16.04, and 1 Gbps public & private uplink in dal13
        slcli order place-quote --name "foobar" --send-email CLOUD_SERVER DALLAS13 \\
            GUEST_CORES_4 \\
            RAM_16_GB \\
            REBOOT_REMOTE_CONSOLE \\
            1_GBPS_PUBLIC_PRIVATE_NETWORK_UPLINKS \\
            BANDWIDTH_0_GB_2 \\
            1_IP_ADDRESS \\
            GUEST_DISK_100_GB_SAN \\
            OS_UBUNTU_16_04_LTS_XENIAL_XERUS_MINIMAL_64_BIT_FOR_VSI \\
            MONITORING_HOST_PING \\
            NOTIFICATION_EMAIL_AND_TICKET \\
            AUTOMATED_NOTIFICATION \\
            UNLIMITED_SSL_VPN_USERS_1_PPTP_VPN_USER_PER_ACCOUNT \\
            --extras '{"virtualGuests": [{"hostname": "test", "domain": "softlayer.com"}]}' \\
            --complex-type SoftLayer_Container_Product_Order_Virtual_Guest

    """
    manager = ordering.OrderingManager(env.client)

    if extras:
        try:
            extras = json.loads(extras)
        except ValueError as err:
            raise exceptions.CLIAbort(f"There was an error when parsing the --extras value: {err}")

    args = (package_keyname, location, order_items)
    kwargs = {'preset_keyname': preset,
              'extras': extras,
              'quantity': 1,
              'quote_name': name,
              'send_email': send_email,
              'complex_type': complex_type}

    order = manager.place_quote(*args, **kwargs)

    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'
    table.add_row(['id', order['quote']['id']])
    table.add_row(['name', order['quote']['name']])
    table.add_row(['created', order['orderDate']])
    table.add_row(['expires', order['quote']['expirationDate']])
    table.add_row(['status', order['quote']['status']])
    env.fout(table)



getFrontendMacAddresses = ['06-00-00-00-00-00']
getBackendMacAddresses = ['07-00-00-00-00-00']
getFrontendMacAddresses = ['06-00-00-00-00-00']
getRouter = 'brc01'
getVlans = [10, 124]
getVlanIds = [8384, 12446]
getDatacenter = 'dal01'
getUserMetadata = 'User-supplied data'



"""Invoice listing"""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.managers.account import AccountManager as AccountManager
from SoftLayer import utils


@click.command(cls=SLCommand)
@click.option('--all', 'get_all', is_flag=True, default=False, show_default=True,
              help="Return ALL invoices. There may be a lot of these.")
@click.option('--closed', is_flag=True, default=False, show_default=True,
              help="Include invoices with a CLOSED status.")
@click.option('--limit', default=50, show_default=True,
              help="How many invoices to get back.")
@environment.pass_env
def cli(env, limit, closed=False, get_all=False):
    """List invoices"""

    manager = AccountManager(env.client)
    invoices = manager.get_invoices(limit, closed, get_all)

    table = formatting.Table([
        "Id", "Created", "Type", "Status", "Starting Balance", "Ending Balance", "Invoice Amount", "Items"
    ])
    table.align['Starting Balance'] = 'l'
    table.align['Ending Balance'] = 'l'
    table.align['Invoice Amount'] = 'l'
    table.align['Items'] = 'l'
    if isinstance(invoices, dict):
        invoices = [invoices]
    for invoice in invoices:
        table.add_row([
            invoice.get('id'),
            utils.clean_time(invoice.get('createDate'), out_format="%Y-%m-%d"),
            invoice.get('typeCode'),
            invoice.get('statusCode'),
            invoice.get('startingBalance'),
            invoice.get('endingBalance'),
            invoice.get('invoiceTotalAmount'),
            invoice.get('itemCount')
        ])
    env.fout(table)



"""Virtual server order options."""
# :license: MIT, see LICENSE for more details.
# pylint: disable=too-many-statements
import click
from SoftLayer.managers import network

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, short_help="Get options to use for creating virtual servers.")
@click.argument('location', required=False)
@click.option('--vsi-type', required=False, show_default=True, default='PUBLIC_CLOUD_SERVER',
              type=click.Choice(['PUBLIC_CLOUD_SERVER', 'TRANSIENT_CLOUD_SERVER', 'SUSPEND_CLOUD_SERVER',
                                 'CLOUD_SERVER']),
              help="VS keyName type.")
@click.option('--prices', '-p', is_flag=True,
              help='Use --prices to list the server item prices, and to list the Item Prices by location,'
                   'add it to the --prices option using location short name, e.g. --prices dal13')
@environment.pass_env
def cli(env, vsi_type, prices, location=None):
    """Virtual server order options."""

    vsi = SoftLayer.VSManager(env.client)
    network_manager = network.NetworkManager(env.client)
    options = vsi.get_create_options(vsi_type, location)

    pods = network_manager.get_closed_pods()

    tables = []

    # Datacenters
    dc_table = formatting.Table(['Datacenter', 'Value', 'Note'], title="Datacenters")
    dc_table.sortby = 'Value'
    dc_table.align = 'l'
    for location_info in options['locations']:
        closure = []
        for pod in pods:
            if location_info['key'] in str(pod['name']):
                closure.append(pod['name'])

        notes = '-'
        if len(closure) > 0:
            notes = f"closed soon: {', '.join(closure)}"
        dc_table.add_row([location_info['name'], location_info['key'], notes])
    tables.append(dc_table)

    if vsi_type == 'CLOUD_SERVER':
        tables.append(guest_core_prices_table(options['guest_core'], prices))
        tables.append(ram_prices_table(options['ram'], prices))
    else:
        tables.append(preset_prices_table(options['sizes'], prices))
    tables.append(os_prices_table(options['operating_systems'], prices))
    tables.append(port_speed_prices_table(options['port_speed'], prices))
    tables.append(database_prices_table(options['database'], prices))
    tables.append(guest_disk_prices_table(options['guest_disk'], prices))
    tables.append(extras_prices_table(options['extras'], prices))

    env.fout(tables)


def preset_prices_table(sizes, prices=False):
    """Shows Server Preset options prices.

    :param [] sizes: List of Hardware Server sizes.
    :param prices: Include pricing information or not.
    """
    preset_price_table = formatting.Table(['Size', 'Value', 'Hourly', 'Monthly'], title="Sizes Prices")
    preset_price_table.sortby = 'Value'
    preset_price_table.align = 'l'

    preset_table = formatting.Table(['Size', 'Value'], title="Sizes")
    preset_table.sortby = 'Value'
    preset_table.align = 'l'

    for size in sizes:
        if (size['hourlyRecurringFee'] > 0) or (size['recurringFee'] > 0):
            preset_price_table.add_row([size['name'], size['key'], "%.4f" % size['hourlyRecurringFee'],
                                        "%.4f" % size['recurringFee']])
        preset_table.add_row([size['name'], size['key']])
    if prices:
        return preset_price_table
    return preset_table


def os_prices_table(operating_systems, prices=False):
    """Shows Server Operating Systems prices cost and capacity restriction.

    :param [] operating_systems: List of Hardware Server operating systems.
    :param prices: Include pricing information or not.
    """
    os_price_table = formatting.Table(['OS Key', 'Hourly', 'Monthly', 'Restriction'],
                                      title="Operating Systems Prices")
    os_price_table.sortby = 'OS Key'
    os_price_table.align = 'l'

    os_table = formatting.Table(['OS', 'Key', 'Reference Code'], title="Operating Systems")
    os_table.sortby = 'Key'
    os_table.align = 'l'

    for operating_system in operating_systems:
        for price in operating_system['prices']:
            cr_max = _get_price_data(price, 'capacityRestrictionMaximum')
            cr_min = _get_price_data(price, 'capacityRestrictionMinimum')
            cr_type = _get_price_data(price, 'capacityRestrictionType')
            os_price_table.add_row(
                [operating_system['key'],
                 _get_price_data(price, 'hourlyRecurringFee'),
                 _get_price_data(price, 'recurringFee'),
                 f"{cr_min} - {cr_max} {cr_type}"])
        os_table.add_row([operating_system['name'], operating_system['key'], operating_system['referenceCode']])
    if prices:
        return os_price_table
    return os_table


def port_speed_prices_table(port_speeds, prices=False):
    """Shows Server Port Speeds prices cost and capacity restriction.

    :param [] port_speeds: List of Hardware Server Port Speeds.
    :param prices: Include pricing information or not.
    """
    port_speed_price_table = formatting.Table(['Key', 'Speed', 'Hourly', 'Monthly'], title="Network Options Prices")
    port_speed_price_table.sortby = 'Speed'
    port_speed_price_table.align = 'l'

    port_speed_table = formatting.Table(['network', 'Key'], title="Network Options")
    port_speed_table.sortby = 'Key'
    port_speed_table.align = 'l'

    for speed in port_speeds:
        for price in speed['prices']:
            port_speed_price_table.add_row(
                [speed['key'], speed['speed'],
                 _get_price_data(price, 'hourlyRecurringFee'),
                 _get_price_data(price, 'recurringFee')])
        port_speed_table.add_row([speed['name'], speed['key']])
    if prices:
        return port_speed_price_table
    return port_speed_table


def extras_prices_table(extras, prices=False):
    """Shows Server extras prices cost and capacity restriction.

    :param [] extras: List of Hardware Server Extras.
    :param prices: Include pricing information or not.
    """
    extras_price_table = formatting.Table(['Extra Option Key', 'Hourly', 'Monthly'], title="Extras Prices")
    extras_price_table.align = 'l'

    extras_table = formatting.Table(['Extra Option', 'Value'], title="Extras")
    extras_table.sortby = 'Value'
    extras_table.align = 'l'

    for extra in extras:
        for price in extra['prices']:
            extras_price_table.add_row(
                [extra['key'],
                 _get_price_data(price, 'hourlyRecurringFee'),
                 _get_price_data(price, 'recurringFee')])
        extras_table.add_row([extra['name'], extra['key']])
    if prices:
        return extras_price_table
    return extras_table


def ram_prices_table(ram_list, prices=False):
    """Shows Server Port Speeds prices cost and capacity restriction.

    :param [] ram_list: List of Virtual Server Ram.
    :param prices: Include pricing information or not.
    """
    ram_price_table = formatting.Table(['Key', 'Hourly', 'Monthly'], title="Ram Prices")
    ram_price_table.sortby = 'Key'
    ram_price_table.align = 'l'

    ram_table = formatting.Table(['memory', 'Value'], title="RAM")
    ram_table.sortby = 'Value'
    ram_table.align = 'l'

    for ram in ram_list:
        for price in ram['prices']:
            ram_price_table.add_row(
                [ram['key'],
                 _get_price_data(price, 'hourlyRecurringFee'),
                 _get_price_data(price, 'recurringFee')])
        ram_table.add_row([ram['name'], ram['key']])
    if prices:
        return ram_price_table
    return ram_table


def database_prices_table(database_list, prices=False):
    """Shows Server Port Speeds prices cost and capacity restriction.

    :param [] database_list: List of Virtual Server database.
    :param prices: Include pricing information or not.
    """
    database_price_table = formatting.Table(['Key', 'Hourly', 'Monthly', 'Restriction'], title="Data Base Prices")
    database_price_table.sortby = 'Key'
    database_price_table.align = 'l'

    database_table = formatting.Table(['database', 'Value'], title="Databases")
    database_table.sortby = 'Value'
    database_table.align = 'l'

    for database in database_list:
        for price in database['prices']:
            cr_max = _get_price_data(price, 'capacityRestrictionMaximum')
            cr_min = _get_price_data(price, 'capacityRestrictionMinimum')
            cr_type = _get_price_data(price, 'capacityRestrictionType')
            database_price_table.add_row(
                [database['key'],
                 _get_price_data(price, 'hourlyRecurringFee'),
                 _get_price_data(price, 'recurringFee'),
                 f"{cr_min} - {cr_max} {cr_type}"])
        database_table.add_row([database['name'], database['key']])
    if prices:
        return database_price_table
    return database_table


def guest_core_prices_table(guest_core_list, prices=False):
    """Shows Server Port Speeds prices cost and capacity restriction.

    :param [] guest_core_list: List of Virtual Server guest_core.
    :param prices: Include pricing information or not.
    """
    guest_core_price_table = formatting.Table(['Key', 'Hourly', 'Monthly'], title="Guest Core Prices")
    guest_core_price_table.sortby = 'Key'
    guest_core_price_table.align = 'l'

    guest_core_table = formatting.Table(['cpu', 'Value', 'Capacity'], title="Guest_core")
    guest_core_table.sortby = 'Value'
    guest_core_table.align = 'l'

    for guest_core in guest_core_list:
        for price in guest_core['prices']:
            guest_core_price_table.add_row(
                [guest_core['key'],
                 _get_price_data(price, 'hourlyRecurringFee'),
                 _get_price_data(price, 'recurringFee')])
        guest_core_table.add_row([guest_core['name'], guest_core['key'], guest_core['capacity']])
    if prices:
        return guest_core_price_table
    return guest_core_table


def guest_disk_prices_table(guest_disk_list, prices=False):
    """Shows Server Port Speeds prices cost and capacity restriction.

    :param [] guest_disk_list: List of Virtual Server guest_disk.
    :param prices: Include pricing information or not.
    """
    guest_disk_price_table = formatting.Table(['Key', 'Hourly', 'Monthly'], title="Guest Disk Prices")
    guest_disk_price_table.sortby = 'Key'
    guest_disk_price_table.align = 'l'

    guest_disk_table = formatting.Table(['guest_disk', 'Value', 'Capacity', 'Disk'], title="Guest_disks")
    guest_disk_table.sortby = 'Value'
    guest_disk_table.align = 'l'

    for guest_disk in guest_disk_list:
        for price in guest_disk['prices']:
            guest_disk_price_table.add_row(
                [guest_disk['key'],
                 _get_price_data(price, 'hourlyRecurringFee'),
                 _get_price_data(price, 'recurringFee')])
        guest_disk_table.add_row(
            [guest_disk['name'], guest_disk['key'], guest_disk['capacity'], guest_disk['disk']])
    if prices:
        return guest_disk_price_table
    return guest_disk_table


def _get_price_data(price, item):
    """Get a specific data from HS price.

    :param price: Hardware Server price.
    :param string item: Hardware Server price data.
    """
    result = '-'
    if item in price:
        result = price[item]
    return result



"""Manage Migrations of Virtual Guests"""
# :license: MIT, see LICENSE for more details.
import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.option('--guest', '-g', type=click.INT, help="Guest ID to immediately migrate.")
@click.option('--all', '-a', 'migrate_all', is_flag=True, default=False,
              help="Migrate ALL guests that require migration immediately.")
@click.option('--host', '-h', type=click.INT,
              help="Dedicated Host ID to migrate to. Only works on guests that are already on a dedicated host.")
@environment.pass_env
def cli(env, guest, migrate_all, host):
    """Manage VSIs that require migration. Can migrate Dedicated Host VSIs as well."""

    vsi = SoftLayer.VSManager(env.client)
    dedicated_filter = {'virtualGuests': {'dedicatedHost': {'id': {'operation': 'not null'}}}}
    mask = """mask[
                id, hostname, domain, datacenter, pendingMigrationFlag, powerState,
                primaryIpAddress,primaryBackendIpAddress, dedicatedHost
            ]"""

    # No options, just print out a list of guests that can be migrated
    if not (guest or migrate_all):
        require_migration = vsi.list_instances(mask=mask)
        require_table = formatting.Table(['id', 'hostname', 'domain', 'datacenter'], title="Require Migration")

        for vsi_object in require_migration:
            if vsi_object.get('pendingMigrationFlag', False):
                require_table.add_row([
                    vsi_object.get('id'),
                    vsi_object.get('hostname'),
                    vsi_object.get('domain'),
                    utils.lookup(vsi_object, 'datacenter', 'name')
                ])

        if len(require_table.rows) > 0:
            env.fout(require_table)
        else:
            click.secho("No guests require migration at this time.", fg='green')

        migrateable = vsi.list_instances(filter=dedicated_filter, mask=mask)
        migrateable_table = formatting.Table(['id', 'hostname', 'domain', 'datacenter', 'Host Name', 'Host Id'],
                                             title="Dedicated Guests")
        for vsi_object in migrateable:
            migrateable_table.add_row([
                vsi_object.get('id'),
                vsi_object.get('hostname'),
                vsi_object.get('domain'),
                utils.lookup(vsi_object, 'datacenter', 'name'),
                utils.lookup(vsi_object, 'dedicatedHost', 'name'),
                utils.lookup(vsi_object, 'dedicatedHost', 'id')
            ])
        if len(migrateable_table.rows) > 0:
            env.fout(migrateable_table)
        else:
            click.secho("No dedicated guests to migrate.", fg='green')
    # Migrate all guests with pendingMigrationFlag=True
    elif migrate_all:
        require_migration = vsi.list_instances(mask="mask[id,pendingMigrationFlag]")
        migrated = 0
        for vsi_object in require_migration:
            if vsi_object.get('pendingMigrationFlag', False):
                migrated = migrated + 1
                migrate(vsi, vsi_object['id'])
        if migrated == 0:
            click.secho("No guests require migration at this time", fg='green')
    # Just migrate based on the options
    else:
        migrate(vsi, guest, host)


def migrate(vsi_manager, vsi_id, host_id=None):
    """Handles actually migrating virtual guests and handling the exception"""

    try:
        if host_id:
            vsi_manager.migrate_dedicated(vsi_id, host_id)
        else:
            vsi_manager.migrate(vsi_id)
        click.secho(f"Started a migration on {vsi_id}", fg='green')
    except SoftLayer.exceptions.SoftLayerAPIError as ex:
        click.secho(f"Failed to migrate {vsi_id}. {str(ex)}", fg='red')



"""Restore a file volume from a snapshot."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('volume_id')
@click.option('--snapshot-id', '-s',
              help='The id of the snapshot which will be used'
              ' to restore the block volume')
@environment.pass_env
def cli(env, volume_id, snapshot_id):
    """Restore file volume using a given snapshot

    Example::
    slcli file snapshot-restore 12345678 -s 87654321
    This command restores volume with ID 12345678 from snapshot with ID 87654321.
"""
    file_manager = SoftLayer.FileStorageManager(env.client)
    success = file_manager.restore_from_snapshot(volume_id, snapshot_id)

    if success:
        click.echo('File volume %s is being restored using snapshot %s'
                   % (volume_id, snapshot_id))



"""List a users permissions."""
import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """User Permissions.

    Some permissions here may also be managed by IBM IAM service.
    See https://cloud.ibm.com/docs/account?topic=account-migrated_permissions for more details.
    """

    mgr = SoftLayer.UserManager(env.client)
    user_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'username')
    object_mask = "mask[id, permissions, isMasterUserFlag, roles]"

    user = mgr.get_user(user_id, object_mask)
    all_permissions = mgr.get_permission_departments()

    user_permissions = perms_to_dict(user['permissions'])
    all_table = formatting.KeyValueTable(['Department', 'Permissions'])
    if user['isMasterUserFlag']:
        click.secho('This account is the Master User and has all permissions enabled', fg='green')

    env.fout(roles_table(user))
    for department in all_permissions:
        all_table.add_row([
            department.get('name'),
            permission_table(user_permissions, department.get('permissions', []))
        ])
    env.fout(all_table)


def perms_to_dict(perms):
    """Takes a list of permissions and transforms it into a dictionary for better searching"""
    permission_dict = {}
    for perm in perms:
        permission_dict[perm['keyName']] = True
    return permission_dict


def permission_table(user_permissions, all_permissions):
    """Creates a table of available permissions"""

    table = formatting.Table(['KeyName', 'Assigned', 'Description'])
    table.align['KeyName'] = 'l'
    table.align['Description'] = 'l'
    table.align['Assigned'] = 'l'
    for perm in all_permissions:
        assigned = user_permissions.get(perm['keyName'], False)
        table.add_row([perm['keyName'], assigned, perm['description']])
    return table


def roles_table(user):
    """Creates a table for a users roles"""
    table = formatting.Table(['id', 'Role Name', 'Description'])
    for role in user['roles']:
        table.add_row([role['id'], role['name'], role['description']])
    return table



"""Get billing for a virtual device."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers
from SoftLayer import utils


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """Get billing for a virtual device."""
    virtual = SoftLayer.VSManager(env.client)

    virtual_id = helpers.resolve_id(virtual.resolve_ids, identifier, 'virtual')
    result = virtual.get_instance(virtual_id)
    table = formatting.KeyValueTable(['name', 'value'])
    table.align['name'] = 'r'
    table.align['value'] = 'l'

    table.add_row(['Id', identifier])

    table.add_row(['Billing Item Id', utils.lookup(result, 'billingItem', 'id')])
    table.add_row(['Recurring Fee', utils.lookup(result, 'billingItem', 'recurringFee')])
    table.add_row(['Total', utils.lookup(result, 'billingItem', 'nextInvoiceTotalRecurringAmount')])
    table.add_row(['Provision Date', utils.lookup(result, 'provisionDate')])

    price_table = formatting.Table(['Description', 'Recurring Price'])
    for item in utils.lookup(result, 'billingItem', 'children') or []:
        price_table.add_row([item['description'], item['nextInvoiceTotalRecurringAmount']])

    table.add_row(['prices', price_table])
    env.fout(table)



"""Enable or Disable specific permissions for a user"""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--enable/--disable', default=True,
              help="Enable (DEFAULT) or Disable selected permissions")
@click.option('--permission', '-p', multiple=True,
              help="Permission keyName to set, multiple instances allowed. "
                   "Use keyword ALL to select ALL permisssions")
@click.option('--from-user', '-u', default=None,
              help="Set permissions to match this user's permissions. "
                   "Will add then remove the appropriate permissions")
@environment.pass_env
def cli(env, identifier, enable, permission, from_user):
    """Enable or Disable specific permissions."""

    mgr = SoftLayer.UserManager(env.client)
    user_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'username')
    result = False
    if from_user:
        from_user_id = helpers.resolve_id(mgr.resolve_ids, from_user, 'username')
        result = mgr.permissions_from_user(user_id, from_user_id)
    elif enable:
        result = mgr.add_permissions(user_id, permission)
    else:
        result = mgr.remove_permissions(user_id, permission)

    if result:
        click.secho(f"Permissions updated successfully: {', '.join(permission)}", fg='green')
    else:
        click.secho(f"Failed to update permissions: {', '.join(permission)}", fg='red')



getObject = {
    'accountId': 1234,
    'address': 'test-01-307608-ams01.clb.appdomain.cloud',
    'createDate': '2019-08-12T07:49:43-06:00',
    'id': 1111111,
    'isPublic': 0,
    'locationId': 265592,
    'modifyDate': '2019-08-13T16:26:06-06:00',
    'name': 'test-01',
    'operatingStatus': 'ONLINE',
    'provisioningStatus': 'ACTIVE',
    'type': 0,
    'useSystemPublicIpPool': 1,
    'uuid': '1a1aa111-4474-4e16-9f02-4de959229b85',
    'listenerCount': 4,
    'memberCount': 1,
    'previousErrorText': 'test',
    'datacenter': {
        'id': 265592,
        'longName': 'Amsterdam 1',
        'name': 'ams01',
        'statusId': 2
    },
    'healthMonitors': [
        {
            'createDate': '2019-08-20T18:05:09-04:00',
            'interval': 5,
            'maxRetries': 2,
            'modifyDate': '2019-08-20T18:05:18-04:00',
            'monitorType': 'HTTP',
            'provisioningStatus': 'ACTIVE',
            'timeout': 2,
            'urlPath': '/',
            'uuid': 'c11111c1-f5ab-4c15-ba96-d7b95dc7c824'
        }
    ],
    'l7Pools': [
        {
            'createDate': '2019-08-19T16:33:37-04:00',
            'id': 222222,
            'loadBalancingAlgorithm': 'ROUNDROBIN',
            'modifyDate': None,
            'name': 'test',
            'protocol': 'HTTP',
            'provisioningStatus': 'ACTIVE',
            'uuid': 'a1111111-c5e7-413f-9f78-84f6c5e1ca04'
        }
    ],
    'listeners': [
        {
            'clientTimeout': 15,
            'defaultPool': {
                'healthMonitor': {
                    'uuid': '222222ab-bbcc-4f32-9b31-1b6d3a1959c8'
                },
                'protocol': 'HTTP',
                'protocolPort': 256,
                'uuid': 'ab1a1abc-0e83-4690-b5d4-1359625dba8f',
            }
        },
        {
            'clientTimeout': 15,
            'defaultPool': {
                'healthMonitor': {
                    'uuid': '222222ab-bbcc-4f32-9b31-1b6d3a1959c0'
                },
                'protocol': 'HTTP',
                'protocolPort': 256,
                'uuid': 'ab1a1abc-0e83-4690-b5d4-1359625dba8x',
            }
        },
        {'connectionLimit': None,
         'createDate': '2019-08-21T17:19:25-04:00',
         'defaultPool': {'createDate': '2019-08-21T17:19:25-04:00',
                         'healthMonitor': {'createDate': '2019-08-21T17:17:04-04:00',
                                           'id': 859330,
                                           'interval': 5,
                                           'maxRetries': 2,
                                           'modifyDate': '2019-08-21T17:17:15-04:00',
                                           'monitorType': 'HTTP',
                                           'provisioningStatus': 'ACTIVE',
                                           'timeout': 2,
                                           'urlPath': '/',
                                           'uuid': '55e00152-75fd-4f94-9263-cb4c6e005f12'},
                         'loadBalancingAlgorithm': 'ROUNDROBIN',
                         'members': [{'address': '10.136.4.220',
                                      'createDate': '2019-08-12T09:49:43-04:00',
                                      'id': 1023118,
                                      'modifyDate': '2019-08-12T09:52:54-04:00',
                                      'provisioningStatus': 'ACTIVE',
                                      'uuid': 'ba23a166-faa4-4eb2-96e7-ef049d65ce60',
                                      'weight': 50}],
                         'modifyDate': '2019-08-21T17:19:33-04:00',
                         'protocol': 'HTTP',
                         'protocolPort': 230,
                         'provisioningStatus': 'ACTIVE',
                         'uuid': '1c5f3ba6-ec7d-4cf8-8815-9bb174224a76'},
         'id': 889072,
         'l7Policies': [{'action': 'REJECT',
                         'createDate': '2019-08-21T18:17:41-04:00',
                         'id': 215204,
                         'modifyDate': None,
                         'name': 'trestst',
                         'priority': 1,
                         'redirectL7PoolId': None,
                         'uuid': 'b8c30aae-3979-49a7-be8c-fb70e43a6b4b'}],
         'modifyDate': '2019-08-22T10:58:02-04:00',
         'protocol': 'HTTP',
         'protocolPort': 110,
         'provisioningStatus': 'ACTIVE',
         'tlsCertificateId': None,
         'clientTimeout': 30,
         'uuid': 'a509723d-a3cb-4ae4-bc5b-5ecf04f890ff'}
    ],
    'members': [
        {
            'address': '10.0.0.1',
            'createDate': '2019-08-12T09:49:43-04:00',
            'modifyDate': '2019-08-12T09:52:54-04:00',
            'provisioningStatus': 'ACTIVE',
            'uuid': 'ba23a166-faa4-4eb2-96e7-ef049d65ce60',
            'weight': 50
        }
    ],
    'sslCiphers': [
        {
            'id': 2, 'name': 'ECDHE-RSA-AES256-GCM-SHA384'
        }
    ],
}
getAllObjects = [getObject]

getLoadBalancer = {
    "accountId": 3071234,
    "createDate": "2019-08-12T21:49:43+08:00",
    "id": 81234,
    "isPublic": 0,
    "locationId": 265592,
    "modifyDate": "2019-08-14T06:26:06+08:00",
    "name": "dcabero-01",
    "uuid": "0a2da082-4474-4e16-9f02-4de11111",
    "datacenter": {
        "id": 265592,
        "longName": "Amsterdam 1",
        "name": "ams01",
        "statusId": 2
    }
}

cancelLoadBalancer = True

getLoadBalancerMemberHealth = [
    {
        'poolUuid': '1c5f3ba6-ec7d-4cf8-8815-9bb174224a76',
        'membersHealth': [
            {
                'status': 'DOWN',
                'uuid': 'ba23a166-faa4-4eb2-96e7-ef049d65ce60'
            }
        ]
    }
]

getHealthMonitors = {}

getLoadBalancer = getObject
cancelLoadBalancer = getObject



"""Order/create a dedicated server."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting
from SoftLayer.CLI import helpers
from SoftLayer.CLI import template


@click.command(cls=SoftLayer.CLI.command.SLCommand, epilog="See 'slcli server create-options' for valid options.")
@click.option('--hostname', '-H', required=True, prompt=True, help="Host portion of the FQDN")
@click.option('--domain', '-D', required=True, prompt=True, help="Domain portion of the FQDN")
@click.option('--size', '-s', required=True, prompt=True, help="Hardware size")
@click.option('--os', '-o', required=True, prompt=True, help="OS Key value")
@click.option('--datacenter', '-d', required=True, prompt=True, help="Datacenter shortname")
@click.option('--port-speed', type=click.INT, help="Port speeds. DEPRECATED, use --network")
@click.option('--no-public', is_flag=True, help="Private network only. DEPRECATED, use --network.")
@click.option('--network', help="Network Option Key. Use instead of port-speed option")
@click.option('--billing', default='hourly', show_default=True, type=click.Choice(['hourly', 'monthly']),
              help="Billing rate")
@click.option('--postinstall', '-i', help="Post-install script. Should be a HTTPS URL.")
@click.option('--test', is_flag=True, help="Do not actually create the server")
@click.option('--template', '-t', is_eager=True, type=click.Path(exists=True, readable=True, resolve_path=True),
              callback=template.TemplateCallback(list_args=['key']),
              help="A template file that defaults the command-line options")
@click.option('--export', type=click.Path(writable=True, resolve_path=True),
              help="Exports options to a template file")
@click.option('--wait', type=click.INT,
              help="Wait until the server is finished provisioning for up to X seconds before returning")
@click.option('--router-public', type=click.INT,
              help="The ID of the public ROUTER on which you want the virtual server placed")
@click.option('--router-private', type=click.INT,
              help="The ID of the private ROUTER on which you want the virtual server placed")
@helpers.multi_option('--key', '-k', help="SSH keys to add to the root user")
@helpers.multi_option('--extra', '-e', help="Extra option Key Names")
@click.option('--force',  default=False, is_flag=True, help="Force modify")
@environment.pass_env
def cli(env, force, **args):
    """Order/create a dedicated server."""
    mgr = SoftLayer.HardwareManager(env.client)
    network = SoftLayer.NetworkManager(env.client)

    pods = network.get_closed_pods()

    # Get the SSH keys
    ssh_keys = []
    for key in args.get('key'):
        resolver = SoftLayer.SshKeyManager(env.client).resolve_ids
        key_id = helpers.resolve_id(resolver, key, 'SshKey')
        ssh_keys.append(key_id)

    order = {
        'hostname': args['hostname'],
        'domain': args['domain'],
        'size': args['size'],
        'location': args.get('datacenter'),
        'ssh_keys': ssh_keys,
        'post_uri': args.get('postinstall'),
        'os': args['os'],
        'hourly': args.get('billing') == 'hourly',
        'port_speed': args.get('port_speed'),
        'no_public': args.get('no_public') or False,
        'extras': args.get('extra'),
        'network': args.get('network'),
        'public_router': args.get('router_public', None),
        'private_router': args.get('router_private', None)
    }

    # Do not create hardware server with --test or --export
    do_create = not (args['export'] or args['test'])

    output = None
    if args.get('test'):
        result = mgr.verify_order(**order)

        table = formatting.Table(['Item', 'cost'])
        table.align['Item'] = 'r'
        table.align['cost'] = 'r'

        total = 0.0
        for price in result['prices']:
            total += float(price.get('recurringFee', 0.0))
            rate = "%.2f" % float(price['recurringFee'])

            table.add_row([price['item']['description'], rate])

        table.add_row(['Total monthly cost', "%.2f" % total])
        output = []
        output.append(table)
        output.append(formatting.FormattedItem(
            '',
            ' -- ! Prices reflected here are retail and do not '
            'take account level discounts and are not guaranteed.'))

    if args['export']:
        export_file = args.pop('export')
        template.export_to_template(export_file, args, exclude=['wait', 'test'])
        env.fout('Successfully exported options to a template file.')
        return

    if do_create:
        for pod in pods:
            if args.get('datacenter') in pod['name']:
                click.secho(f"Warning: Closed soon: {pod['name']}", fg='yellow')
        if not force:
            if not (env.skip_confirmations or formatting.confirm(
                    "This action will incur charges on your account. Continue?")):
                raise exceptions.CLIAbort('Aborting dedicated server order.')

        result = mgr.place_order(**order)

        table = formatting.KeyValueTable(['name', 'value'])
        table.align['name'] = 'r'
        table.align['value'] = 'l'
        table.add_row(['id', result['orderId']])
        table.add_row(['created', result['orderDate']])
        output = table

    env.fout(output)



"""Modify an existing file storage volume."""
# :license: MIT, see LICENSE for more details.

import click
import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting


CONTEXT_SETTINGS = {'token_normalize_func': lambda x: x.upper()}


@click.command(cls=SoftLayer.CLI.command.SLCommand, context_settings=CONTEXT_SETTINGS)
@click.argument('volume-id')
@click.option('--new-size', '-c',
              type=int,
              help='New Size of file volume in GB. ***If no size is given, the original size of volume is used.***\n'
                   'Potential Sizes: [20, 40, 80, 100, 250, 500, 1000, 2000, 4000, 8000, 12000]\n'
                   'Minimum: [the original size of the volume]')
@click.option('--new-iops', '-i',
              type=int,
              help='Performance Storage IOPS, between 100 and 6000 in multiples of 100 [only for performance volumes] '
                   '***If no IOPS value is specified, the original IOPS value of the volume will be used.***')
@click.option('--new-tier', '-t',
              help='Endurance Storage Tier (IOPS per GB) [only for endurance volumes] '
                   '***If no tier is specified, the original tier of the volume will be used.***',
              type=click.Choice(['0.25', '2', '4', '10']))
@click.option('--force',  default=False, is_flag=True, help="Force modify")
@environment.pass_env
def cli(env, volume_id, new_size, new_iops, new_tier, force):
    """Modify an existing file storage volume.

    Valid size and iops options can be found here:
        https://cloud.ibm.com/docs/FileStorage/index.html#provisioning-considerations
        https://cloud.ibm.com/docs/FileStorage?topic=FileStorage-orderingFileStorage&interface=cli

    Example::

            slcli file volume-modify 12345678 --new-size 1000 --new-iops 400
    """
    if not force:
        if not (env.skip_confirmations or
                formatting.confirm("This action will incur charges on your account. Continue?")):
            raise exceptions.CLIAbort('Aborted')

    file_manager = SoftLayer.FileStorageManager(env.client)

    if new_tier is not None:
        new_tier = float(new_tier)

    try:
        order = file_manager.order_modified_volume(
            volume_id,
            new_size=new_size,
            new_iops=new_iops,
            new_tier_level=new_tier,
        )
    except ValueError as ex:
        raise exceptions.ArgumentError(str(ex))

    if 'placedOrder' in order.keys():
        click.echo(f"Order #{order['placedOrder']['id']} placed successfully!")
        for item in order['placedOrder']['items']:
            click.echo(" > %s" % item['description'])
    else:
        click.echo("Order could not be placed! Please verify your options and try again.")



"""
    SoftLayer.CLI.routes
    ~~~~~~~~~~~~~~~~~~~~~
    This is how all commands are registered with the CLI.

    :license: MIT, see LICENSE for more details.
"""

ALL_ROUTES = [
    ('shell', 'SoftLayer.shell.core:cli'),
    ('emplogin', 'SoftLayer.CLI.login:cli'),

    ('call-api', 'SoftLayer.CLI.call_api:cli'),

    ('account', 'SoftLayer.CLI.account'),
    ('account:invoice-detail', 'SoftLayer.CLI.account.invoice_detail:cli'),
    ('account:invoices', 'SoftLayer.CLI.account.invoices:cli'),
    ('account:events', 'SoftLayer.CLI.account.events:cli'),
    ('account:bandwidth-pools-detail', 'SoftLayer.CLI.bandwidth.pools_detail:cli'),
    ('account:event-detail', 'SoftLayer.CLI.account.event_detail:cli'),
    ('account:licenses', 'SoftLayer.CLI.account.licenses:cli'),
    ('account:summary', 'SoftLayer.CLI.account.summary:cli'),
    ('account:billing-items', 'SoftLayer.CLI.account.billing_items:cli'),
    ('account:item-detail', 'SoftLayer.CLI.account.item_detail:cli'),
    ('account:cancel-item', 'SoftLayer.CLI.account.cancel_item:cli'),
    ('account:orders', 'SoftLayer.CLI.account.orders:cli'),
    ('account:bandwidth-pools', 'SoftLayer.CLI.bandwidth.pools:cli'),
    ('account:hooks', 'SoftLayer.CLI.account.hooks:cli'),
    ('account:hook-create', 'SoftLayer.CLI.account.hook_create:cli'),
    ('account:hook-delete', 'SoftLayer.CLI.account.hook_delete:cli'),

    ('virtual', 'SoftLayer.CLI.virt'),
    ('virtual:bandwidth', 'SoftLayer.CLI.virt.bandwidth:cli'),
    ('virtual:billing', 'SoftLayer.CLI.virt.billing:cli'),
    ('virtual:cancel', 'SoftLayer.CLI.virt.cancel:cli'),
    ('virtual:capture', 'SoftLayer.CLI.virt.capture:cli'),
    ('virtual:create', 'SoftLayer.CLI.virt.create:cli'),
    ('virtual:create-options', 'SoftLayer.CLI.virt.create_options:cli'),
    ('virtual:detail', 'SoftLayer.CLI.virt.detail:cli'),
    ('virtual:dns-sync', 'SoftLayer.CLI.virt.dns:cli'),
    ('virtual:edit', 'SoftLayer.CLI.virt.edit:cli'),
    ('virtual:list', 'SoftLayer.CLI.virt.list:cli'),
    ('virtual:pause', 'SoftLayer.CLI.virt.power:pause'),
    ('virtual:power-off', 'SoftLayer.CLI.virt.power:power_off'),
    ('virtual:power-on', 'SoftLayer.CLI.virt.power:power_on'),
    ('virtual:rescue', 'SoftLayer.CLI.virt.power:rescue'),
    ('virtual:resume', 'SoftLayer.CLI.virt.power:resume'),
    ('virtual:ready', 'SoftLayer.CLI.virt.ready:cli'),
    ('virtual:reboot', 'SoftLayer.CLI.virt.power:reboot'),
    ('virtual:reload', 'SoftLayer.CLI.virt.reload:cli'),
    ('virtual:storage', 'SoftLayer.CLI.virt.storage:cli'),
    ('virtual:upgrade', 'SoftLayer.CLI.virt.upgrade:cli'),
    ('virtual:usage', 'SoftLayer.CLI.virt.usage:cli'),
    ('virtual:credentials', 'SoftLayer.CLI.virt.credentials:cli'),
    ('virtual:authorize-storage', 'SoftLayer.CLI.virt.authorize_storage:cli'),
    ('virtual:capacity', 'SoftLayer.CLI.virt.capacity:cli'),
    ('virtual:placementgroup', 'SoftLayer.CLI.virt.placementgroup:cli'),
    ('virtual:migrate', 'SoftLayer.CLI.virt.migrate:cli'),
    ('virtual:access', 'SoftLayer.CLI.virt.access:cli'),
    ('virtual:monitoring', 'SoftLayer.CLI.virt.monitoring:cli'),
    ('virtual:notifications', 'SoftLayer.CLI.virt.notifications:cli'),
    ('virtual:notification-add', 'SoftLayer.CLI.virt.notification_add:cli'),
    ('virtual:notification-delete', 'SoftLayer.CLI.virt.notification_delete:cli'),
    ('virtual:host-list', 'SoftLayer.CLI.dedicatedhost.list:cli'),
    ('virtual:host-create', 'SoftLayer.CLI.dedicatedhost.create:cli'),
    ('virtual:os-available', 'SoftLayer.CLI.virt.os_available:cli'),

    ('dedicatedhost', 'SoftLayer.CLI.dedicatedhost'),
    ('dedicatedhost:list', 'SoftLayer.CLI.dedicatedhost.list:cli'),
    ('dedicatedhost:create', 'SoftLayer.CLI.dedicatedhost.create:cli'),
    ('dedicatedhost:create-options', 'SoftLayer.CLI.dedicatedhost.create_options:cli'),
    ('dedicatedhost:detail', 'SoftLayer.CLI.dedicatedhost.detail:cli'),
    ('dedicatedhost:cancel', 'SoftLayer.CLI.dedicatedhost.cancel:cli'),
    ('dedicatedhost:cancel-guests', 'SoftLayer.CLI.dedicatedhost.cancel_guests:cli'),
    ('dedicatedhost:list-guests', 'SoftLayer.CLI.dedicatedhost.list_guests:cli'),

    ('cdn', 'SoftLayer.CLI.cdn'),
    ('cdn:detail', 'SoftLayer.CLI.cdn.cdn:cli'),
    ('cdn:edit', 'SoftLayer.CLI.cdn.cdn:cli'),
    ('cdn:list', 'SoftLayer.CLI.cdn.cdn:cli'),
    ('cdn:origin-add', 'SoftLayer.CLI.cdn.cdn:cli'),
    ('cdn:origin-list', 'SoftLayer.CLI.cdn.cdn:cli'),
    ('cdn:origin-remove', 'SoftLayer.CLI.cdn.cdn:cli'),
    ('cdn:purge', 'SoftLayer.CLI.cdn.cdn:cli'),
    ('cdn:delete', 'SoftLayer.CLI.cdn.cdn:cli'),
    ('cdn:create', 'SoftLayer.CLI.cdn.cdn:cli'),

    ('config', 'SoftLayer.CLI.config'),
    ('config:setup', 'SoftLayer.CLI.config.setup:cli'),
    ('config:show', 'SoftLayer.CLI.config.show:cli'),
    ('setup', 'SoftLayer.CLI.config.setup:cli'),

    ('dns', 'SoftLayer.CLI.dns'),
    ('dns:import', 'SoftLayer.CLI.dns.zone_import:cli'),
    ('dns:record-add', 'SoftLayer.CLI.dns.record_add:cli'),
    ('dns:record-edit', 'SoftLayer.CLI.dns.record_edit:cli'),
    ('dns:record-list', 'SoftLayer.CLI.dns.record_list:cli'),
    ('dns:record-remove', 'SoftLayer.CLI.dns.record_remove:cli'),
    ('dns:zone-create', 'SoftLayer.CLI.dns.zone_create:cli'),
    ('dns:zone-delete', 'SoftLayer.CLI.dns.zone_delete:cli'),
    ('dns:zone-list', 'SoftLayer.CLI.dns.zone_list:cli'),
    ('dns:zone-print', 'SoftLayer.CLI.dns.zone_print:cli'),

    ('block', 'SoftLayer.CLI.block'),
    ('block:access-authorize', 'SoftLayer.CLI.block.access.authorize:cli'),
    ('block:access-list', 'SoftLayer.CLI.block.access.list:cli'),
    ('block:access-revoke', 'SoftLayer.CLI.block.access.revoke:cli'),
    ('block:access-password', 'SoftLayer.CLI.block.access.password:cli'),
    ('block:duplicate-convert-status', 'SoftLayer.CLI.block.duplicate_convert_status:cli'),
    ('block:subnets-list', 'SoftLayer.CLI.block.subnets.list:cli'),
    ('block:subnets-assign', 'SoftLayer.CLI.block.subnets.assign:cli'),
    ('block:subnets-remove', 'SoftLayer.CLI.block.subnets.remove:cli'),
    ('block:replica-failback', 'SoftLayer.CLI.block.replication.failback:cli'),
    ('block:replica-failover', 'SoftLayer.CLI.block.replication.failover:cli'),
    ('block:disaster-recovery-failover', 'SoftLayer.CLI.block.replication.disaster_recovery_failover:cli'),
    ('block:replica-order', 'SoftLayer.CLI.block.replication.order:cli'),
    ('block:replica-partners', 'SoftLayer.CLI.block.replication.partners:cli'),
    ('block:replica-locations', 'SoftLayer.CLI.block.replication.locations:cli'),
    ('block:snapshot-cancel', 'SoftLayer.CLI.block.snapshot.cancel:cli'),
    ('block:snapshot-create', 'SoftLayer.CLI.block.snapshot.create:cli'),
    ('block:snapshot-delete', 'SoftLayer.CLI.block.snapshot.delete:cli'),
    ('block:snapshot-disable', 'SoftLayer.CLI.block.snapshot.disable:cli'),
    ('block:snapshot-set-notification', 'SoftLayer.CLI.block.snapshot.set_notify_status:cli'),
    ('block:snapshot-get-notification-status', 'SoftLayer.CLI.block.snapshot.get_notify_status:cli'),
    ('block:snapshot-enable', 'SoftLayer.CLI.block.snapshot.enable:cli'),
    ('block:snapshot-schedule-list', 'SoftLayer.CLI.block.snapshot.schedule_list:cli'),
    ('block:snapshot-list', 'SoftLayer.CLI.block.snapshot.list:cli'),
    ('block:snapshot-order', 'SoftLayer.CLI.block.snapshot.order:cli'),
    ('block:snapshot-restore', 'SoftLayer.CLI.block.snapshot.restore:cli'),
    ('block:volume-cancel', 'SoftLayer.CLI.block.cancel:cli'),
    ('block:volume-count', 'SoftLayer.CLI.block.count:cli'),
    ('block:volume-detail', 'SoftLayer.CLI.block.detail:cli'),
    ('block:volume-duplicate', 'SoftLayer.CLI.block.duplicate:cli'),
    ('block:volume-list', 'SoftLayer.CLI.block.list:cli'),
    ('block:volume-modify', 'SoftLayer.CLI.block.modify:cli'),
    ('block:volume-order', 'SoftLayer.CLI.block.order:cli'),
    ('block:volume-set-lun-id', 'SoftLayer.CLI.block.lun:cli'),
    ('block:volume-limits', 'SoftLayer.CLI.block.limit:cli'),
    ('block:volume-refresh', 'SoftLayer.CLI.block.refresh:cli'),
    ('block:volume-convert', 'SoftLayer.CLI.block.convert:cli'),
    ('block:volume-options', 'SoftLayer.CLI.block.options:cli'),
    ('block:volume-set-note', 'SoftLayer.CLI.block.set_note:cli'),
    ('block:object-list', 'SoftLayer.CLI.block.object_list:cli'),
    ('block:object-storage-detail', 'SoftLayer.CLI.block.object_storage_detail:cli'),
    ('block:object-storage-permission', 'SoftLayer.CLI.block.object_storage_permission:cli'),

    ('email', 'SoftLayer.CLI.email'),
    ('email:list', 'SoftLayer.CLI.email.list:cli'),
    ('email:detail', 'SoftLayer.CLI.email.detail:cli'),
    ('email:edit', 'SoftLayer.CLI.email.edit:cli'),

    ('licenses', 'SoftLayer.CLI.licenses'),
    ('licenses:create-options', 'SoftLayer.CLI.licenses.create_options:cli'),

    ('event-log', 'SoftLayer.CLI.event_log'),
    ('event-log:get', 'SoftLayer.CLI.event_log.get:cli'),
    ('event-log:types', 'SoftLayer.CLI.event_log.types:cli'),

    ('file', 'SoftLayer.CLI.file'),
    ('file:access-authorize', 'SoftLayer.CLI.file.access.authorize:cli'),
    ('file:access-list', 'SoftLayer.CLI.file.access.list:cli'),
    ('file:access-revoke', 'SoftLayer.CLI.file.access.revoke:cli'),
    ('file:duplicate-convert-status', 'SoftLayer.CLI.file.duplicate_convert_status:cli'),
    ('file:replica-failback', 'SoftLayer.CLI.file.replication.failback:cli'),
    ('file:replica-failover', 'SoftLayer.CLI.file.replication.failover:cli'),
    ('file:disaster-recovery-failover', 'SoftLayer.CLI.file.replication.disaster_recovery_failover:cli'),
    ('file:replica-order', 'SoftLayer.CLI.file.replication.order:cli'),
    ('file:replica-partners', 'SoftLayer.CLI.file.replication.partners:cli'),
    ('file:replica-locations', 'SoftLayer.CLI.file.replication.locations:cli'),
    ('file:snapshot-cancel', 'SoftLayer.CLI.file.snapshot.cancel:cli'),
    ('file:snapshot-create', 'SoftLayer.CLI.file.snapshot.create:cli'),
    ('file:snapshot-delete', 'SoftLayer.CLI.file.snapshot.delete:cli'),
    ('file:snapshot-disable', 'SoftLayer.CLI.file.snapshot.disable:cli'),
    ('file:snapshot-enable', 'SoftLayer.CLI.file.snapshot.enable:cli'),
    ('file:snapshot-set-notification', 'SoftLayer.CLI.file.snapshot.set_notify_status:cli'),
    ('file:snapshot-get-notification-status', 'SoftLayer.CLI.file.snapshot.get_notify_status:cli'),
    ('file:snapshot-schedule-list', 'SoftLayer.CLI.file.snapshot.schedule_list:cli'),
    ('file:snapshot-list', 'SoftLayer.CLI.file.snapshot.list:cli'),
    ('file:snapshot-order', 'SoftLayer.CLI.file.snapshot.order:cli'),
    ('file:snapshot-restore', 'SoftLayer.CLI.file.snapshot.restore:cli'),
    ('file:volume-cancel', 'SoftLayer.CLI.file.cancel:cli'),
    ('file:volume-count', 'SoftLayer.CLI.file.count:cli'),
    ('file:volume-detail', 'SoftLayer.CLI.file.detail:cli'),
    ('file:volume-duplicate', 'SoftLayer.CLI.file.duplicate:cli'),
    ('file:volume-list', 'SoftLayer.CLI.file.list:cli'),
    ('file:volume-modify', 'SoftLayer.CLI.file.modify:cli'),
    ('file:volume-order', 'SoftLayer.CLI.file.order:cli'),
    ('file:volume-limits', 'SoftLayer.CLI.file.limit:cli'),
    ('file:volume-refresh', 'SoftLayer.CLI.file.refresh:cli'),
    ('file:volume-convert', 'SoftLayer.CLI.file.convert:cli'),
    ('file:volume-options', 'SoftLayer.CLI.file.options:cli'),
    ('file:volume-set-note', 'SoftLayer.CLI.file.set_note:cli'),

    ('firewall', 'SoftLayer.CLI.firewall'),
    ('firewall:add', 'SoftLayer.CLI.firewall.add:cli'),
    ('firewall:cancel', 'SoftLayer.CLI.firewall.cancel:cli'),
    ('firewall:detail', 'SoftLayer.CLI.firewall.detail:cli'),
    ('firewall:edit', 'SoftLayer.CLI.firewall.edit:cli'),
    ('firewall:list', 'SoftLayer.CLI.firewall.list:cli'),
    ('firewall:monitoring', 'SoftLayer.CLI.firewall.monitoring:cli'),

    ('globalip', 'SoftLayer.CLI.globalip'),
    ('globalip:assign', 'SoftLayer.CLI.globalip.assign:cli'),
    ('globalip:cancel', 'SoftLayer.CLI.globalip.cancel:cli'),
    ('globalip:create', 'SoftLayer.CLI.globalip.create:cli'),
    ('globalip:list', 'SoftLayer.CLI.globalip.list:cli'),
    ('globalip:unassign', 'SoftLayer.CLI.globalip.unassign:cli'),

    ('image', 'SoftLayer.CLI.image'),
    ('image:delete', 'SoftLayer.CLI.image.delete:cli'),
    ('image:detail', 'SoftLayer.CLI.image.detail:cli'),
    ('image:edit', 'SoftLayer.CLI.image.edit:cli'),
    ('image:list', 'SoftLayer.CLI.image.list:cli'),
    ('image:import', 'SoftLayer.CLI.image.import:cli'),
    ('image:export', 'SoftLayer.CLI.image.export:cli'),
    ('image:datacenter', 'SoftLayer.CLI.image.datacenter:cli'),
    ('image:share', 'SoftLayer.CLI.image.share:cli'),
    ('image:share-deny', 'SoftLayer.CLI.image.share_deny:cli'),

    ('loadbal', 'SoftLayer.CLI.loadbal'),
    ('loadbal:detail', 'SoftLayer.CLI.loadbal.detail:cli'),
    ('loadbal:list', 'SoftLayer.CLI.loadbal.list:cli'),
    ('loadbal:health', 'SoftLayer.CLI.loadbal.health:cli'),
    ('loadbal:member-add', 'SoftLayer.CLI.loadbal.members:add'),
    ('loadbal:member-del', 'SoftLayer.CLI.loadbal.members:remove'),
    ('loadbal:l7policies', 'SoftLayer.CLI.loadbal.layer7_policy_list:policies'),
    ('loadbal:pool-add', 'SoftLayer.CLI.loadbal.pools:add'),
    ('loadbal:pool-edit', 'SoftLayer.CLI.loadbal.pools:edit'),
    ('loadbal:pool-del', 'SoftLayer.CLI.loadbal.pools:delete'),
    ('loadbal:l7pool-add', 'SoftLayer.CLI.loadbal.pools:l7pool_add'),
    ('loadbal:l7pool-del', 'SoftLayer.CLI.loadbal.pools:l7pool_del'),
    ('loadbal:order', 'SoftLayer.CLI.loadbal.order:order'),
    ('loadbal:order-options', 'SoftLayer.CLI.loadbal.order:order_options'),
    ('loadbal:cancel', 'SoftLayer.CLI.loadbal.order:cancel'),
    ('loadbal:protocol-add', 'SoftLayer.CLI.loadbal.protocol_add:cli'),
    ('loadbal:protocol-edit', 'SoftLayer.CLI.loadbal.protocol_edit:cli'),
    ('loadbal:protocol-delete', 'SoftLayer.CLI.loadbal.protocol_delete:cli'),

    ('loadbal:ns-detail', 'SoftLayer.CLI.loadbal.ns_detail:cli'),
    ('loadbal:ns-list', 'SoftLayer.CLI.loadbal.ns_list:cli'),

    ('metadata', 'SoftLayer.CLI.metadata:cli'),

    ('nas', 'SoftLayer.CLI.nas'),
    ('nas:list', 'SoftLayer.CLI.nas.list:cli'),
    ('nas:credentials', 'SoftLayer.CLI.nas.credentials:cli'),

    ('licenses', 'SoftLayer.CLI.licenses'),
    ('licenses:create', 'SoftLayer.CLI.licenses.create:cli'),
    ('licenses:cancel', 'SoftLayer.CLI.licenses.cancel:cli'),

    ('object-storage', 'SoftLayer.CLI.object_storage'),

    ('object-storage:accounts', 'SoftLayer.CLI.object_storage.list_accounts:cli'),
    ('object-storage:endpoints', 'SoftLayer.CLI.object_storage.list_endpoints:cli'),
    ('object-storage:credential', 'SoftLayer.CLI.object_storage.credential:cli'),
    ('object-storage:cancel', 'SoftLayer.CLI.object_storage.cancel:cli'),

    ('order', 'SoftLayer.CLI.order'),
    ('order:category-list', 'SoftLayer.CLI.order.category_list:cli'),
    ('order:item-list', 'SoftLayer.CLI.order.item_list:cli'),
    ('order:package-list', 'SoftLayer.CLI.order.package_list:cli'),
    ('order:place', 'SoftLayer.CLI.order.place:cli'),
    ('order:preset-list', 'SoftLayer.CLI.order.preset_list:cli'),
    ('order:package-locations', 'SoftLayer.CLI.order.package_locations:cli'),
    ('order:place-quote', 'SoftLayer.CLI.order.place_quote:cli'),
    ('order:quote-list', 'SoftLayer.CLI.order.quote_list:cli'),
    ('order:quote-detail', 'SoftLayer.CLI.order.quote_detail:cli'),
    ('order:quote-delete', 'SoftLayer.CLI.order.quote_delete:cli'),
    ('order:quote-save', 'SoftLayer.CLI.order.quote_save:cli'),
    ('order:quote', 'SoftLayer.CLI.order.quote:cli'),
    ('order:lookup', 'SoftLayer.CLI.order.lookup:cli'),
    ('order:cancelation', 'SoftLayer.CLI.order.cancelation:cli'),

    ('hardware', 'SoftLayer.CLI.hardware'),
    ('hardware:bandwidth', 'SoftLayer.CLI.hardware.bandwidth:cli'),
    ('hardware:cancel', 'SoftLayer.CLI.hardware.cancel:cli'),
    ('hardware:cancel-reasons', 'SoftLayer.CLI.hardware.cancel_reasons:cli'),
    ('hardware:create', 'SoftLayer.CLI.hardware.create:cli'),
    ('hardware:create-options', 'SoftLayer.CLI.hardware.create_options:cli'),
    ('hardware:detail', 'SoftLayer.CLI.hardware.detail:cli'),
    ('hardware:billing', 'SoftLayer.CLI.hardware.billing:cli'),
    ('hardware:edit', 'SoftLayer.CLI.hardware.edit:cli'),
    ('hardware:list', 'SoftLayer.CLI.hardware.list:cli'),
    ('hardware:power-cycle', 'SoftLayer.CLI.hardware.power:power_cycle'),
    ('hardware:power-off', 'SoftLayer.CLI.hardware.power:power_off'),
    ('hardware:power-on', 'SoftLayer.CLI.hardware.power:power_on'),
    ('hardware:reboot', 'SoftLayer.CLI.hardware.power:reboot'),
    ('hardware:reload', 'SoftLayer.CLI.hardware.reload:cli'),
    ('hardware:credentials', 'SoftLayer.CLI.hardware.credentials:cli'),
    ('hardware:update-firmware', 'SoftLayer.CLI.hardware.update_firmware:cli'),
    ('hardware:reflash-firmware', 'SoftLayer.CLI.hardware.reflash_firmware:cli'),
    ('hardware:rescue', 'SoftLayer.CLI.hardware.power:rescue'),
    ('hardware:ready', 'SoftLayer.CLI.hardware.ready:cli'),
    ('hardware:toggle-ipmi', 'SoftLayer.CLI.hardware.toggle_ipmi:cli'),
    ('hardware:authorize-storage', 'SoftLayer.CLI.hardware.authorize_storage:cli'),
    ('hardware:dns-sync', 'SoftLayer.CLI.hardware.dns:cli'),
    ('hardware:storage', 'SoftLayer.CLI.hardware.storage:cli'),
    ('hardware:upgrade', 'SoftLayer.CLI.hardware.upgrade:cli'),
    ('hardware:sensor', 'SoftLayer.CLI.hardware.sensor:cli'),
    ('hardware:monitoring', 'SoftLayer.CLI.hardware.monitoring:cli'),
    ('hardware:notifications', 'SoftLayer.CLI.hardware.notifications:cli'),
    ('hardware:notification-add', 'SoftLayer.CLI.hardware.notification_add:cli'),
    ('hardware:notification-delete', 'SoftLayer.CLI.hardware.notification_delete:cli'),
    ('hardware:create-credential', 'SoftLayer.CLI.hardware.create_credential:cli'),
    ('hardware:vlan-trunkable', 'SoftLayer.CLI.hardware.vlan_trunkable:cli'),
    ('hardware:vlan-add', 'SoftLayer.CLI.hardware.vlan_add:cli'),
    ('hardware:vlan-remove', 'SoftLayer.CLI.hardware.vlan_remove:cli'),

    ('securitygroup', 'SoftLayer.CLI.securitygroup'),
    ('securitygroup:list', 'SoftLayer.CLI.securitygroup.list:cli'),
    ('securitygroup:detail', 'SoftLayer.CLI.securitygroup.detail:cli'),
    ('securitygroup:create', 'SoftLayer.CLI.securitygroup.create:cli'),
    ('securitygroup:edit', 'SoftLayer.CLI.securitygroup.edit:cli'),
    ('securitygroup:delete', 'SoftLayer.CLI.securitygroup.delete:cli'),
    ('securitygroup:rule-list', 'SoftLayer.CLI.securitygroup.rule:rule_list'),
    ('securitygroup:rule-add', 'SoftLayer.CLI.securitygroup.rule:add'),
    ('securitygroup:rule-edit', 'SoftLayer.CLI.securitygroup.rule:edit'),
    ('securitygroup:rule-remove', 'SoftLayer.CLI.securitygroup.rule:remove'),
    ('securitygroup:interface-list',
     'SoftLayer.CLI.securitygroup.interface:interface_list'),
    ('securitygroup:interface-add',
     'SoftLayer.CLI.securitygroup.interface:add'),
    ('securitygroup:interface-remove',
     'SoftLayer.CLI.securitygroup.interface:remove'),
    ('securitygroup:event-log', 'SoftLayer.CLI.securitygroup.event_log:get_by_request_id'),

    ('sshkey', 'SoftLayer.CLI.sshkey'),
    ('sshkey:add', 'SoftLayer.CLI.security.sshkey_add:cli'),
    ('sshkey:remove', 'SoftLayer.CLI.security.sshkey_remove:cli'),
    ('sshkey:edit', 'SoftLayer.CLI.security.sshkey_edit:cli'),
    ('sshkey:list', 'SoftLayer.CLI.security.sshkey_list:cli'),
    ('sshkey:print', 'SoftLayer.CLI.security.sshkey_print:cli'),

    ('ssl', 'SoftLayer.CLI.ssl'),
    ('ssl:add', 'SoftLayer.CLI.security.cert_add:cli'),
    ('ssl:download', 'SoftLayer.CLI.security.cert_download:cli'),
    ('ssl:edit', 'SoftLayer.CLI.security.cert_edit:cli'),
    ('ssl:list', 'SoftLayer.CLI.security.cert_list:cli'),
    ('ssl:remove', 'SoftLayer.CLI.security.cert_remove:cli'),

    ('security', 'SoftLayer.CLI.security'),
    ('security:sshkey-add', 'SoftLayer.CLI.security.sshkey_add:cli'),
    ('security:sshkey-remove', 'SoftLayer.CLI.security.sshkey_remove:cli'),
    ('security:sshkey-edit', 'SoftLayer.CLI.security.sshkey_edit:cli'),
    ('security:sshkey-list', 'SoftLayer.CLI.security.sshkey_list:cli'),
    ('security:sshkey-print', 'SoftLayer.CLI.security.sshkey_print:cli'),
    ('security:cert-add', 'SoftLayer.CLI.security.cert_add:cli'),
    ('security:cert-download', 'SoftLayer.CLI.security.cert_download:cli'),
    ('security:cert-edit', 'SoftLayer.CLI.security.cert_edit:cli'),
    ('security:cert-list', 'SoftLayer.CLI.security.cert_list:cli'),
    ('security:cert-remove', 'SoftLayer.CLI.security.cert_remove:cli'),

    ('subnet', 'SoftLayer.CLI.subnet'),
    ('subnet:cancel', 'SoftLayer.CLI.subnet.cancel:cli'),
    ('subnet:create', 'SoftLayer.CLI.subnet.create:cli'),
    ('subnet:edit', 'SoftLayer.CLI.subnet.edit:cli'),
    ('subnet:detail', 'SoftLayer.CLI.subnet.detail:cli'),
    ('subnet:list', 'SoftLayer.CLI.subnet.list:cli'),
    ('subnet:lookup', 'SoftLayer.CLI.subnet.lookup:cli'),
    ('subnet:edit-ip', 'SoftLayer.CLI.subnet.edit_ip:cli'),
    ('subnet:route', 'SoftLayer.CLI.subnet.route:cli'),
    ('subnet:clear-route', 'SoftLayer.CLI.subnet.clear_route:cli'),

    ('tags', 'SoftLayer.CLI.tags'),
    ('tags:cleanup', 'SoftLayer.CLI.tags.cleanup:cli'),
    ('tags:list', 'SoftLayer.CLI.tags.list:cli'),
    ('tags:set', 'SoftLayer.CLI.tags.set:cli'),
    ('tags:details', 'SoftLayer.CLI.tags.details:cli'),
    ('tags:delete', 'SoftLayer.CLI.tags.delete:cli'),
    ('tags:taggable', 'SoftLayer.CLI.tags.taggable:cli'),

    ('ticket', 'SoftLayer.CLI.ticket'),
    ('ticket:create', 'SoftLayer.CLI.ticket.create:cli'),
    ('ticket:detail', 'SoftLayer.CLI.ticket.detail:cli'),
    ('ticket:list', 'SoftLayer.CLI.ticket.list:cli'),
    ('ticket:update', 'SoftLayer.CLI.ticket.update:cli'),
    ('ticket:upload', 'SoftLayer.CLI.ticket.upload:cli'),
    ('ticket:subjects', 'SoftLayer.CLI.ticket.subjects:cli'),
    ('ticket:summary', 'SoftLayer.CLI.ticket.summary:cli'),
    ('ticket:attach', 'SoftLayer.CLI.ticket.attach:cli'),
    ('ticket:detach', 'SoftLayer.CLI.ticket.detach:cli'),

    ('user', 'SoftLayer.CLI.user'),
    ('user:list', 'SoftLayer.CLI.user.list:cli'),
    ('user:detail', 'SoftLayer.CLI.user.detail:cli'),
    ('user:permissions', 'SoftLayer.CLI.user.permissions:cli'),
    ('user:edit-permissions', 'SoftLayer.CLI.user.edit_permissions:cli'),
    ('user:notifications', 'SoftLayer.CLI.user.notifications:cli'),
    ('user:edit-notifications', 'SoftLayer.CLI.user.edit_notifications:cli'),
    ('user:edit-details', 'SoftLayer.CLI.user.edit_details:cli'),
    ('user:create', 'SoftLayer.CLI.user.create:cli'),
    ('user:delete', 'SoftLayer.CLI.user.delete:cli'),
    ('user:device-access', 'SoftLayer.CLI.user.device_access:cli'),
    ('user:vpn-manual', 'SoftLayer.CLI.user.vpn_manual:cli'),
    ('user:vpn-subnet', 'SoftLayer.CLI.user.vpn_subnet:cli'),
    ('user:vpn-enable', 'SoftLayer.CLI.user.vpn_enable_or_disable:vpn_enable'),
    ('user:vpn-disable', 'SoftLayer.CLI.user.vpn_enable_or_disable:vpn_disable'),
    ('user:remove-access', 'SoftLayer.CLI.user.remove_access:cli'),
    ('user:grant-access', 'SoftLayer.CLI.user.grant_access:cli'),
    ('user:vpn-password', 'SoftLayer.CLI.user.vpn_password:cli'),
    ('user:apikey', 'SoftLayer.CLI.user.apikey:cli'),

    ('vlan', 'SoftLayer.CLI.vlan'),
    ('vlan:create', 'SoftLayer.CLI.vlan.create:cli'),
    ('vlan:create-options', 'SoftLayer.CLI.vlan.create_options:cli'),
    ('vlan:detail', 'SoftLayer.CLI.vlan.detail:cli'),
    ('vlan:edit', 'SoftLayer.CLI.vlan.edit:cli'),
    ('vlan:list', 'SoftLayer.CLI.vlan.list:cli'),
    ('vlan:cancel', 'SoftLayer.CLI.vlan.cancel:cli'),

    ('summary', 'SoftLayer.CLI.summary:cli'),
    ('search', 'SoftLayer.CLI.search:cli'),

    ('report', 'SoftLayer.CLI.report'),
    ('report:bandwidth', 'SoftLayer.CLI.bandwidth.summary:cli'),
    ('report:datacenter-closures', 'SoftLayer.CLI.report.dc_closures:cli'),

    ('bandwidth', 'SoftLayer.CLI.bandwidth'),
    ('bandwidth:summary', 'SoftLayer.CLI.bandwidth.summary:cli'),
    ('bandwidth:pools', 'SoftLayer.CLI.bandwidth.pools:cli'),
    ('bandwidth:pools-detail', 'SoftLayer.CLI.bandwidth.pools_detail:cli'),
    ('bandwidth:pools-create', 'SoftLayer.CLI.bandwidth.pools_create:cli'),
    ('bandwidth:pools-edit', 'SoftLayer.CLI.bandwidth.pools_edit:cli'),
    ('bandwidth:pools-delete', 'SoftLayer.CLI.bandwidth.pools_delete:cli'),
]

ALL_ALIASES = {
    'hw': 'hardware',
    'lb': 'loadbal',
    'meta': 'metadata',
    'my': 'metadata',
    'sg': 'securitygroup',
    'server': 'hardware',
    'vm': 'virtual',
    'vs': 'virtual',
    'dh': 'dedicatedhost',
    'pg': 'placementgroup',
}



"""
    SoftLayer.image
    ~~~~~~~~~~~~~~~
    Image Manager/helpers

    :license: MIT, see LICENSE for more details.
"""

from SoftLayer import exceptions
from SoftLayer import utils

IMAGE_MASK = ('id,accountId,name,globalIdentifier,blockDevices,parentId,'
              'createDate,transaction')


class ImageManager(utils.IdentifierMixin, object):
    """Manages SoftLayer server images.

    See product information here:
    https://cloud.ibm.com/docs/image-templates

    :param SoftLayer.API.BaseClient client: the client instance
    """

    def __init__(self, client):
        self.client = client
        self.vgbdtg = self.client['Virtual_Guest_Block_Device_Template_Group']
        self.resolvers = [self._get_ids_from_name_public,
                          self._get_ids_from_name_private]

    def get_image(self, image_id, **kwargs):
        """Get details about an image.

        :param int image: The ID of the image.
        :param dict \\*\\*kwargs: response-level options (mask, limit, etc.)
        """
        if 'mask' not in kwargs:
            kwargs['mask'] = IMAGE_MASK

        return self.vgbdtg.getObject(id=image_id, **kwargs)

    def delete_image(self, image_id):
        """Deletes the specified image.

        :param int image_id: The ID of the image.
        """
        self.vgbdtg.deleteObject(id=image_id)

    def list_private_images(self, guid=None, name=None, limit=100, **kwargs):
        """List all private images.

        :param string guid: filter based on GUID
        :param string name: filter based on name
        :param dict \\*\\*kwargs: response-level options (mask, limit, etc.)
        """
        if 'mask' not in kwargs:
            kwargs['mask'] = IMAGE_MASK

        _filter = utils.NestedDict(kwargs.get('filter') or {})
        _filter['privateBlockDeviceTemplateGroups']['id'] = utils.query_filter_orderby()
        if name:
            _filter['privateBlockDeviceTemplateGroups']['name'] = utils.query_filter(name)

        if guid:
            _filter['privateBlockDeviceTemplateGroups']['globalIdentifier'] = utils.query_filter(guid)

        kwargs['filter'] = _filter.to_dict()

        return self.client.iter_call('SoftLayer_Account', 'getPrivateBlockDeviceTemplateGroups',
                                     filter=kwargs['filter'], mask=kwargs['mask'], limit=limit)

    def list_public_images(self, guid=None, name=None, limit=100, **kwargs):
        """List all public images.

        :param string guid: filter based on GUID
        :param string name: filter based on name
        :param dict \\*\\*kwargs: response-level options (mask, limit, etc.)
        """
        if 'mask' not in kwargs:
            kwargs['mask'] = IMAGE_MASK

        _filter = utils.NestedDict(kwargs.get('filter') or {})
        _filter['id'] = utils.query_filter_orderby()
        if name:
            _filter['name'] = utils.query_filter(name)

        if guid:
            _filter['globalIdentifier'] = utils.query_filter(guid)

        kwargs['filter'] = _filter.to_dict()

        return self.client.iter_call('SoftLayer_Virtual_Guest_Block_Device_Template_Group', 'getPublicImages',
                                     filter=kwargs['filter'], mask=kwargs['mask'], limit=limit)

    def _get_ids_from_name_public(self, name):
        """Get public images which match the given name."""
        results = self.list_public_images(name=name)
        return [result['id'] for result in results]

    def _get_ids_from_name_private(self, name):
        """Get private images which match the given name."""
        results = self.list_private_images(name=name)
        return [result['id'] for result in results]

    def edit(self, image_id, name=None, note=None, tag=None):
        """Edit image related details.

        :param int image_id: The ID of the image
        :param string name: Name of the Image.
        :param string note: Note of the image.
        :param string tag: Tags of the image to be updated to.
        """
        obj = {}
        if name:
            obj['name'] = name
        if note:
            obj['note'] = note
        if obj:
            self.vgbdtg.editObject(obj, id=image_id)
        if tag:
            self.vgbdtg.setTags(str(tag), id=image_id)

        return bool(name or note or tag)

    def import_image_from_uri(self, name, uri, os_code=None, note=None,
                              ibm_api_key=None, root_key_crn=None,
                              wrapped_dek=None, cloud_init=False,
                              byol=False, is_encrypted=False):
        """Import a new image from object storage.

        :param string name: Name of the new image
        :param string uri: The URI for an object storage object
            (.vhd/.iso file) of the format:
            swift://<objectStorageAccount>@<cluster>/<container>/<objectPath>
            or (.vhd/.iso/.raw file) of the format:
            cos://<regionName>/<bucketName>/<objectPath> if using IBM Cloud
            Object Storage
        :param string os_code: The reference code of the operating system
        :param string note: Note to add to the image
        :param string ibm_api_key: Ibm Api Key needed to communicate with ICOS
            and your KMS
        :param string root_key_crn: CRN of the root key in your KMS. Go to your
            KMS (Key Protect or Hyper Protect) provider to get the CRN for your
            root key.  An example CRN:
            crn:v1:bluemix:public:hs-crypto:us-south:acctID:serviceID:key:keyID'
            Used only when is_encrypted is True.
        :param string wrapped_dek: Wrapped Data Encryption Key provided by
            your KMS. Used only when is_encrypted is True.
        :param boolean cloud_init: Specifies if image is cloud-init
        :param boolean byol: Specifies if image is bring your own license
        :param boolean is_encrypted: Specifies if image is encrypted
        """
        if 'cos://' in uri:
            return self.vgbdtg.createFromIcos({
                'name': name,
                'note': note,
                'operatingSystemReferenceCode': os_code,
                'uri': uri,
                'ibmApiKey': ibm_api_key,
                'crkCrn': root_key_crn,
                'wrappedDek': wrapped_dek,
                'cloudInit': cloud_init,
                'byol': byol,
                'isEncrypted': is_encrypted
            })
        else:
            return self.vgbdtg.createFromExternalSource({
                'name': name,
                'note': note,
                'operatingSystemReferenceCode': os_code,
                'uri': uri,
            })

    def export_image_to_uri(self, image_id, uri, ibm_api_key=None):
        """Export image into the given object storage

        :param int image_id: The ID of the image
        :param string uri: The URI for object storage of the format
            swift://<objectStorageAccount>@<cluster>/<container>/<objectPath>
            or cos://<regionName>/<bucketName>/<objectPath> if using IBM Cloud
            Object Storage
        :param string ibm_api_key: Ibm Api Key needed to communicate with IBM
            Cloud Object Storage
        """
        if 'cos://' in uri:
            return self.vgbdtg.copyToIcos({
                'uri': uri,
                'ibmApiKey': ibm_api_key
            }, id=image_id)
        else:
            return self.vgbdtg.copyToExternalSource({'uri': uri}, id=image_id)

    def add_locations(self, image_id, location_names):
        """Add available locations to an archive image template.

        :param int image_id: The ID of the image
        :param location_names: Locations for the Image.
        """
        locations = self.get_locations_list(image_id, location_names)
        return self.vgbdtg.addLocations(locations, id=image_id)

    def remove_locations(self, image_id, location_names):
        """Remove available locations from an archive image template.

        :param int image_id: The ID of the image
        :param location_names: Locations for the Image.
        """
        locations = self.get_locations_list(image_id, location_names)
        return self.vgbdtg.removeLocations(locations, id=image_id)

    def get_storage_locations(self, image_id):
        """Get available locations for public image storage.

        :param int image_id: The ID of the image
        """
        return self.vgbdtg.getStorageLocations(id=image_id)

    def get_locations_list(self, image_id, location_names):
        """Converts a list of location names to a list of locations.

        :param int image_id: The ID of the image.
        :param list location_names: A list of location names strings.
        :returns: A list of locations associated with the given location names in the image.
        """
        locations = self.get_storage_locations(image_id)
        locations_ids = []
        matching_location = {}
        output_error = "Location {} does not exist for available locations for image {}"

        for location_name in location_names:
            for location in locations:
                if location_name == location.get('name'):
                    matching_location = location
                    break
            if matching_location.get('id') is None:
                raise exceptions.SoftLayerError(output_error.format(location_name, image_id))

            locations_ids.append(matching_location)

        return locations_ids

    def share_image(self, image_id, account_id):
        """Permit sharing image template with another account.

        :param int image_id: The ID of the image.
        :param int account_id: The ID of the another account to share the image.
        """

        return self.vgbdtg.permitSharingAccess(account_id, id=image_id)

    def deny_share_image(self, image_id, account_id):
        """Deny sharing image template with another account.

        :param int image_id: The ID of the image.
        :param int account_id: The ID of the another account to deny share the image.
        """

        return self.vgbdtg.denySharingAccess(account_id, id=image_id)



"""Assigns the global IP to a target."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import helpers


# pylint: disable=unused-argument
def targetipcallback(ctx, param, value):
    """This is here to allow for using --target-id in some cases. Takes the first value and returns it"""
    if value:
        return value[0]
    return value


@click.command(cls=SoftLayer.CLI.command.SLCommand, epilog="More information about types and identifiers "
               "on https://sldn.softlayer.com/reference/services/SoftLayer_Network_Subnet/route/")
@click.argument('globalip')
@click.argument('targetip', nargs=-1, callback=targetipcallback)
@click.option('--target', type=click.Choice(['vlan', 'ip', 'hardware', 'vsi']),
              help='choose the type. vlan, ip, hardware, vsi')
@click.option('--target-id', help='The identifier for the destination resource to route this subnet to.')
@environment.pass_env
def cli(env, globalip, targetip, target, target_id):
    """Assigns the GLOBALIP to TARGETIP.

    GLOBALIP should be either the Global IP address, or the SoftLayer_Network_Subnet_IpAddress_Global id
    See `slcli globalip list`
    TARGETIP should be either the target IP address, or the SoftLayer_Network_Subnet_IpAddress id
    See `slcli subnet list`
    Example::

        slcli globalip assign 12345678 9.111.123.456
        This command assigns Global IP address with ID 12345678 to a target device whose IP address is 9.111.123.456

        slcli globalip assign 123.4.5.6 6.5.4.123
        Global IPs can be specified by their IP address
    """
    mgr = SoftLayer.NetworkManager(env.client)
    # Find SoftLayer_Network_Subnet_IpAddress_Global::id
    global_ip_id = helpers.resolve_id(mgr.resolve_global_ip_ids, globalip,  name='Global IP')

    # Find Global IPs SoftLayer_Network_Subnet::id
    mask = "mask[id,ipAddress[subnetId]]"
    subnet = env.client.call('SoftLayer_Network_Subnet_IpAddress_Global', 'getObject', id=global_ip_id, mask=mask)
    subnet_id = subnet.get('ipAddress', {}).get('subnetId')

    # For backwards compatibility
    if target_id:
        targetip = target_id

    mgr.route(subnet_id, 'SoftLayer_Network_Subnet_IpAddress', targetip)



"""Edits an SSH key."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--label', '-k', help="The new label for the key")
@click.option('--note', help="New notes for the key")
@environment.pass_env
def cli(env, identifier, label, note):
    """Edits an SSH key."""

    mgr = SoftLayer.SshKeyManager(env.client)

    key_id = helpers.resolve_id(mgr.resolve_ids, identifier, 'SshKey')

    if not mgr.edit_key(key_id, label=label, notes=note):
        raise exceptions.CLIAbort('Failed to edit SSH key')



"""Edit details of an image."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import helpers


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--name', help="Name of the image")
@click.option('--note', help="Additional note for the image")
@click.option('--tag', help="Tags for the image")
@environment.pass_env
def cli(env, identifier, name, note, tag):
    """Edit details of an image."""

    image_mgr = SoftLayer.ImageManager(env.client)
    data = {}
    if name:
        data['name'] = name
    if note:
        data['note'] = note
    if tag:
        data['tag'] = tag
    image_id = helpers.resolve_id(image_mgr.resolve_ids, identifier, 'image')
    if not image_mgr.edit(image_id, **data):
        raise exceptions.CLIAbort("Failed to Edit Image")



"""
    SoftLayer.decoration
    ~~~~~~~~~~~~~~~~~~~~
    Handy decorators to use

    :license: MIT, see LICENSE for more details.
"""
from functools import wraps
from random import randint
from time import sleep

from SoftLayer import exceptions

RETRIABLE = (
    exceptions.ServerError,
    exceptions.ApplicationError,
    exceptions.RemoteSystemError,
)


def retry(ex=RETRIABLE, tries=4, delay=5, backoff=2, logger=None):
    """Retry calling the decorated function using an exponential backoff.

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    :param ex: the exception to check. may be a tuple of exceptions to check
    :param tries: number of times to try (not retry) before giving up
    :param delay: initial delay between retries in seconds.
        A random 0-5s will be added to this number to stagger calls.
    :param backoff: backoff multiplier e.g. value of 2 will double the delay each retry
    :param logger: logger to use. If None, print
    """
    def deco_retry(func):
        """@retry(arg[, ...]) -> true decorator"""

        @wraps(func)
        def f_retry(*args, **kwargs):
            """true decorator -> decorated function"""
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return func(*args, **kwargs)
                except ex as error:
                    sleeping = mdelay + randint(0, 5)
                    msg = "%s, Retrying in %d seconds..." % (str(error), sleeping)
                    if logger:
                        logger.warning(msg)
                    sleep(sleeping)
                    mtries -= 1
                    mdelay *= backoff
            return func(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry



"""List NAS account credentials."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@environment.pass_env
def cli(env, identifier):
    """List NAS account credentials."""

    nw_mgr = SoftLayer.NetworkManager(env.client)
    result = nw_mgr.get_nas_credentials(identifier)
    table = formatting.Table(['username', 'password'])
    table.add_row([result.get('username', 'None'),
                   result.get('password', 'None')])
    env.fout(table)



"""Create a user hardware notification entry."""
# :license: MIT, see LICENSE for more details.

import click

import SoftLayer
from SoftLayer.CLI import environment
from SoftLayer.CLI import exceptions
from SoftLayer.CLI import formatting


@click.command(cls=SoftLayer.CLI.command.SLCommand, )
@click.argument('identifier')
@click.option('--users', multiple=True, required=True,
              help='UserId to be notified on monitoring failure.')
@environment.pass_env
def cli(env, identifier, users):
    """Create a user hardware notification entry."""

    hardware = SoftLayer.HardwareManager(env.client)

    table = formatting.KeyValueTable(['Id', 'Hostmane', 'Username', 'Email', 'FirstName', 'Lastname'])
    table.align['Id'] = 'r'
    table.align['Username'] = 'l'

    for user in users:
        notification = hardware.add_notification(identifier, user)
        if notification:
            table.add_row([notification['id'], notification['hardware']['fullyQualifiedDomainName'],
                           notification['user']['username'], notification['user']['email'],
                           notification['user']['firstName'], notification['user']['lastName']])
        else:
            raise exceptions.CLIAbort(f"User not found: {user}.")
    env.fout(table)



getObject = {
    "certificate": "-----BEGIN CERTIFICATE----- \nMIIEJTCCAw2gAwIBAgIDCbQ0MA0GCSqGSIb3DQEBCwUAMEcxCzAJBgNVBAYTAlVT"
                   " -----END CERTIFICATE-----",
    "certificateSigningRequest": "-----BEGIN CERTIFICATE REQUEST-----\n"
                                 "MIIC1jCCAb4CAQAwgZAxCzAJBgNVBAYTAlVTMQ4wDAYDVQQIDAVUZXhh123456QMA4G\n"
                                 "-----END CERTIFICATE REQUEST-----",
    "commonName": "techbabble.xyz",
    "createDate": "2016-01-20T10:56:44-06:00",
    "id": 123456,
    "intermediateCertificate": "",
    "keySize": 258369,
    "modifyDate": "2019-06-05T14:10:40-06:00",
    "organizationName": "Unspecified",
    "privateKey": "-----BEGIN RSA PRIVATE KEY-----\nMIIEpAIBAAKCAQEA3SwTZ7sh7we5zIbmtSbxGJxff07eutrK12345678WXtwQSdE\n"
                  "-----END RSA PRIVATE KEY-----",
    "validityBegin": "2016-01-19T17:59:37-06:00",
    "validityDays": 0,
    "validityEnd": "2017-01-20T13:48:41-06:00"
}
createObject = {}
editObject = True
deleteObject = True



"""Summary info about tickets."""
# :license: MIT, see LICENSE for more details.

import click

from SoftLayer.CLI.command import SLCommand as SLCommand
from SoftLayer.CLI import environment
from SoftLayer.CLI import formatting


@click.command(cls=SLCommand)
@environment.pass_env
def cli(env):
    """Summary info about tickets."""

    mask = ('openTicketCount, closedTicketCount, '
            'openBillingTicketCount, openOtherTicketCount, '
            'openSalesTicketCount, openSupportTicketCount, '
            'openAccountingTicketCount')
    account = env.client['Account'].getObject(mask=mask)
    table = formatting.Table(['Status', 'count'])

    nested = formatting.Table(['Type', 'count'])
    nested.add_row(['Accounting',
                    account['openAccountingTicketCount']])
    nested.add_row(['Billing', account['openBillingTicketCount']])
    nested.add_row(['Sales', account['openSalesTicketCount']])
    nested.add_row(['Support', account['openSupportTicketCount']])
    nested.add_row(['Other', account['openOtherTicketCount']])
    nested.add_row(['Total', account['openTicketCount']])
    table.add_row(['Open', nested])
    table.add_row(['Closed', account['closedTicketCount']])

    env.fout(table)
